# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""Interface for getting Crash and Bucket data from CrashManager API"""
from __future__ import annotations

import json
from contextlib import contextmanager
from logging import getLogger
from pathlib import Path
from re import search
from shutil import copyfileobj, rmtree
from tempfile import NamedTemporaryFile, mkdtemp
from typing import Any, Dict, Generator, cast
from zipfile import BadZipFile, ZipFile

from Collector.Collector import Collector
from FTB.ProgramConfiguration import ProgramConfiguration
from FTB.Signatures.CrashInfo import CrashInfo

from .reporter import Quality
from .storage import TEST_INFO
from .utils import grz_tmp

FM_CONFIG = Path.home() / ".fuzzmanagerconf"
LOG = getLogger(__name__)


class CrashEntry:
    """Get the CrashEntry data for the specified CrashManager crash.

    Attributes:
        crash_id: the server ID for the crash
        see crashmanager.serializers.CrashEntrySerializer
    """

    RAW_FIELDS = frozenset({"rawCrashData", "rawStderr", "rawStdout"})

    __slots__ = (
        "_crash_id",
        "_coll",
        "_contents",
        "_data",
        "_storage",
        "_sig_filename",
        "_url",
    )

    def __init__(self, crash_id: int) -> None:
        """Initialize CrashEntry.

        Arguments:
            crash_id: ID of the requested crash on the server side
        """
        assert isinstance(crash_id, int)
        self._crash_id = crash_id
        self._coll = Collector()
        self._contents: list[Path] | None = None
        self._data: dict[str, Any] | None = None
        self._storage: Path | None = None
        self._sig_filename: Path | None = None
        self._url = (
            f"{self._coll.serverProtocol}://{self._coll.serverHost}:"
            f"{self._coll.serverPort}/crashmanager/rest/crashes/{crash_id}/"
        )

    @property
    def crash_id(self) -> int:
        return self._crash_id

    def __enter__(self) -> CrashEntry:
        return self

    def __exit__(self, *exc: object) -> None:
        self.cleanup()

    def __getattr__(self, name: str) -> Any:
        if self._data is None or (name in self.RAW_FIELDS and name not in self._data):
            need_raw = "1" if name in self.RAW_FIELDS else "0"
            # TODO: handle 403 and 404?
            self._data = cast(
                Dict[str, Any],
                self._coll.get(self._url, params={"include_raw": need_raw}).json(),
            )
        if name not in self._data:
            raise AttributeError(
                f"'{type(self).__name__}' object has no attribute '{name}' "
                f"(has: {list(self._data)})"
            )
        return self._data[name]

    def __setattr__(self, name: str, value: Any) -> None:
        if name.startswith("_"):
            super().__setattr__(name, value)
            return
        if name != "testcase_quality":
            raise AttributeError("can't set attribute")
        self._coll.patch(self._url, data={name: value})
        if self._data:
            self._data[name] = value

    def cleanup(self) -> None:
        """Cleanup any resources held by this instance.

        Arguments:
            None

        Returns:
            None
        """
        if self._storage is not None:
            rmtree(self._storage, ignore_errors=True)
        if self._sig_filename is not None:
            rmtree(self._sig_filename.parent)

    @staticmethod
    def _subset(tests: list[Path], subset: list[int]) -> list[Path]:
        """Select a subset of tests directories. Subset values are sanitized to
        avoid raising.

        Arguments:
            tests: Directories on disk where testcases exists.
            subset: Indices of corresponding directories to select.

        Returns:
            Directories that have been selected.
        """
        assert isinstance(subset, list)
        assert tests
        count = len(tests)
        # deduplicate and limit requested indices to valid range
        keep = {max(count + x, 0) if x < 0 else min(x, count - 1) for x in subset}
        LOG.debug("using TestCase(s) with index %r", keep)
        # build list of items to preserve
        return [tests[i] for i in sorted(keep)]

    def testcases(
        self, subset: list[int] | None = None, ext: str | None = None
    ) -> list[Path]:
        """Download the testcase data from CrashManager.

        Arguments:
            subset: Indices of corresponding directories to select.
            ext: Overwrite file extension used with downloaded testcase.

        Returns:
            Directories on disk where testcases exists.
        """
        if self._contents is None:
            assert self._storage is None
            response = self._coll.get(f"{self._url}download/")

            if "content-disposition" not in response.headers:
                raise RuntimeError(
                    f"Server sent malformed response: {response!r}"
                )  # pragma: no cover

            with NamedTemporaryFile(
                dir=grz_tmp("fuzzmanager"),
                prefix=f"crash-{self.crash_id}-",
                suffix=Path(self.testcase).suffix,
            ) as data:
                data.write(response.content)
                data.seek(0)
                # self._storage should be removed when self.cleanup() is called
                self._storage = Path(
                    mkdtemp(
                        prefix=f"crash-{self.crash_id}-", dir=grz_tmp("fuzzmanager")
                    )
                )
                try:
                    with ZipFile(data) as zip_fp:
                        zip_fp.extractall(path=self._storage)
                    # test case directories are named sequentially
                    # an archive with three test directories would have:
                    # - 'foo-2' (oldest)
                    # - 'foo-1'
                    # - 'foo-0' (most recent)
                    # see FuzzManagerReporter for more info
                    self._contents = sorted(
                        (x.parent for x in self._storage.rglob(TEST_INFO)),
                        reverse=True,
                    )
                except BadZipFile as exc:
                    LOG.debug("downloaded data is not a valid zip")
                    if ext is None:
                        match = search(
                            r'filename="(?P<name>.+)"',
                            response.headers["content-disposition"],
                        )
                        if match:
                            # if nothing is found fallback to html
                            file_name = match.group("name")
                            if "." in file_name:
                                ext = file_name.split(".")[-1] or "html"
                            else:
                                ext = "html"
                    assert ext is not None
                    if ext.lower().endswith("zip"):
                        LOG.error("Error loading test case: %s", exc)
                        self._contents = []
                    else:
                        # load raw test case
                        test_file = self._storage / f"test.{ext}"
                        data.seek(0)
                        with test_file.open("wb") as dst:
                            copyfileobj(data, dst)
                        self._contents = [test_file]

        if subset and self._contents:
            return self._subset(self._contents, subset)
        return self._contents

    def create_signature(self, binary: Path) -> Path:
        """Create a CrashManager signature from this crash.
        If self.bucket is set, self.bucket.signature_path() should be used instead.

        Arguments:
            binary: binary location, needed to create a program configuration

        Returns:
            Signature file.
        """
        if self._sig_filename is not None:
            return self._sig_filename

        success = False
        tmpd = Path(
            mkdtemp(prefix=f"crash-sig-{self._crash_id}-", dir=grz_tmp("fuzzmanager"))
        )
        try:
            sig_basename = f"{self._crash_id}.signature"
            sig_filename = tmpd / sig_basename

            cfg = ProgramConfiguration.fromBinary(binary)
            fm_crash = CrashInfo.fromRawCrashData(
                self.rawStdout, self.rawStderr, cfg, auxCrashData=self.rawCrashData
            )
            fm_sig = fm_crash.createCrashSignature()
            if fm_sig is None:
                raise RuntimeError(str(fm_crash.failureReason))
            sig_filename.write_text(fm_sig.rawSignature)

            sigmeta_filename = sig_filename.with_suffix(".metadata")
            sigmeta_filename.write_text(
                json.dumps(
                    {
                        "size": 1,
                        "frequent": False,
                        "shortDescription": fm_crash.createShortSignature(),
                        "testcase__quality": Quality.UNREDUCED,
                    }
                )
            )
            success = True
        finally:  # pragma: no cover
            if not success:
                rmtree(tmpd)

        self._sig_filename = sig_filename
        return self._sig_filename


class Bucket:
    """Get Bucket data for a specified CrashManager bucket."""

    def __init__(self, bucket_id: int) -> None:
        """Initialize a Bucket instance.

        Arguments:
            bucket_id: ID of the requested bucket on the server.
        """
        assert isinstance(bucket_id, int)
        self._bucket_id = bucket_id
        self._sig_filename: Path | None = None
        self._coll = Collector()
        self._url = (
            f"{self._coll.serverProtocol}://{self._coll.serverHost}:"
            f"{self._coll.serverPort}/crashmanager/rest/buckets/{bucket_id}/"
        )
        self._data: dict[str, Any] | None = None

    @property
    def bucket_id(self) -> int:
        return self._bucket_id

    def __enter__(self) -> Bucket:
        return self

    def __exit__(self, *exc: object) -> None:
        self.cleanup()

    def __getattr__(self, name: str) -> Any:
        if self._data is None:
            self._data = self._coll.get(self._url).json()
        if name not in self._data:
            raise AttributeError(
                f"'{type(self).__name__}' object has no attribute '{name}'"
                f" (has: {list(self._data)})"
            )
        return self._data[name]

    def __setattr__(self, name: str, value: Any) -> None:
        if name.startswith("_"):
            super().__setattr__(name, value)
            return
        raise AttributeError("can't set attribute")

    def cleanup(self) -> None:
        """Cleanup any resources held by this instance.

        Arguments:
            None

        Returns:
            None
        """
        if self._sig_filename is not None:
            rmtree(self._sig_filename.parent)

    def iter_crashes(
        self, quality_filter: int | None = None
    ) -> Generator[CrashEntry, None, None]:
        """Fetch all crash IDs for this FuzzManager bucket.
        Only crashes with testcases are returned.

        Arguments:
            quality_filter: Filter crashes by quality value (None for all)

        Yields:
            CrashEntry objects.
        """

        def _get_results(
            endpoint: str, params: dict[str, str] | None = None
        ) -> Generator[dict[str, Any], None, None]:
            """
            Function to get paginated results from FuzzManager

            Args:
                endpoint: FuzzManager REST API to query (eg. "crashes").
                params: Params to pass through to requests.get.

            Returns:
                Objects (dict) returned by FuzzManager.
            """
            LOG.debug("first request to /%s/", endpoint)

            url = (
                f"{self._coll.serverProtocol}://{self._coll.serverHost}:"
                f"{self._coll.serverPort}/crashmanager/rest/{endpoint}/"
            )

            response: dict[str, Any] = self._coll.get(url, params=params).json()

            while True:
                LOG.debug(
                    "got %d/%d %s",
                    len(response["results"]),
                    response["count"],
                    endpoint,
                )
                while response["results"]:
                    yield response["results"].pop()

                if response["next"] is None:
                    break

                LOG.debug("next request to /%s/", endpoint)
                response = self._coll.get(response["next"]).json()

        # Get all crashes for bucket
        query_args = [
            ("op", "AND"),
            ("bucket", self.bucket_id),
        ]
        if quality_filter is not None:
            query_args.append(("testcase__quality", quality_filter))
        query = json.dumps(dict(query_args))

        n_yielded = 0
        for crash in _get_results(
            "crashes", params={"query": query, "include_raw": "0"}
        ):
            if not crash["testcase"]:
                LOG.warning("crash %d has no testcase, skipping", crash["id"])
                continue

            n_yielded += 1
            LOG.debug("yielding crash #%d", n_yielded)
            result = CrashEntry(cast(int, crash["id"]))
            result._data = crash  # pylint: disable=protected-access
            yield result

    def signature_path(self) -> Path:
        """Download the bucket data from CrashManager.

        Arguments:
            None

        Returns:
            Signature file.
        """
        if self._sig_filename is not None:
            return self._sig_filename

        success = False
        tmpd = Path(
            mkdtemp(prefix=f"bucket-{self._bucket_id}-", dir=grz_tmp("fuzzmanager"))
        )
        try:
            sig_basename = f"{self._bucket_id}.signature"
            sig_filename = tmpd / sig_basename
            sig_filename.write_text(self.signature)
            sigmeta_filename = sig_filename.with_suffix(".metadata")
            sigmeta_filename.write_text(
                json.dumps(
                    {
                        "size": self.size,
                        "frequent": self.frequent,
                        "shortDescription": self.shortDescription,
                        "testcase__quality": self.best_quality,
                    }
                )
            )
            success = True
        finally:  # pragma: no cover
            if not success:
                rmtree(tmpd)

        self._sig_filename = sig_filename
        return self._sig_filename


@contextmanager
def load_fm_data(
    crash_id: int, load_bucket: bool = False
) -> Generator[tuple[CrashEntry, Bucket | None], None, None]:
    """Load CrashEntry including Bucket from FuzzManager.

    Arguments:
        crash_id: Crash ID to load.
        load_bucket: Attempt to load bucket.

    Yields:
        Data loaded from FuzzManager.
    """
    with CrashEntry(crash_id) as crash:
        # load signature if needed
        if load_bucket and crash.bucket:
            with Bucket(crash.bucket) as bucket:
                yield crash, bucket
        else:
            yield crash, None



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""Grizzly reducer strategy definitions.

Each class defined here is an iterator yielding *potential* reductions. The caller
should evaluate each set of testcases, and keep the best one. The caller is responsible
for cleaning up all testcases that are yielded.

Constants:
    DEFAULT_STRATEGIES: Strategy names run by default if unspecified.
    STRATEGIES: Mapping of available strategy names to implementing class.
"""
from __future__ import annotations

from abc import ABC, abstractmethod
from hashlib import sha512
from logging import DEBUG, getLogger
from pathlib import Path
from shutil import rmtree
from tempfile import mkdtemp
from typing import TYPE_CHECKING, Iterable, Iterator, Type, cast

from ...common.utils import grz_tmp

if TYPE_CHECKING:
    from ...common.storage import TestCase

try:
    from pkg_resources import iter_entry_points
except ImportError:
    from ...common.utils import iter_entry_points  # type: ignore

LOG = getLogger(__name__)


DEFAULT_STRATEGIES = (
    "list",
    "lines",
    "cssbeautify",
    "jsbeautify",
    "collapsebraces",
    "attrs",
)


class Strategy(ABC):
    """A strategy is a procedure for repeatedly running a testcase to find the smallest
    equivalent test.

    Implementers must define these class attributes:

    Class Attributes:
        name: The strategy name.
    """

    name: str

    def __init__(self, testcases: list[TestCase]) -> None:
        """Initialize strategy instance.

        Arguments:
            testcases: Testcases to reduce. The object does not take ownership of the
                       testcases.
        """
        # Tuple[str, bytes] is str(path.relative_to(tc_root)) -> hash(testfile data)
        #   for each test file in tc_root (see calculate_testcase_hash below)
        # Tuple[above, ...] is all the test files for a set of test cases meaning:
        #   0/**, 1/**, etc.
        # Set[above] is the unique test case sets which have been tried
        self._tried: set[tuple[tuple[str, bytes], ...]] = set()
        self._testcase_root = Path(mkdtemp(prefix="tc_", dir=grz_tmp("reduce")))
        self.dump_testcases(testcases)

    def _calculate_testcase_hash(self) -> tuple[tuple[str, bytes], ...]:
        """Calculate hashes of all files in testcase root.

        Returns:
            Mapping of file path to SHA-512 of each file in testcase root.
        """
        result: list[tuple[str, bytes]] = []
        for path in self._testcase_root.glob("**/*"):
            if path.is_file():
                tf_hash = sha512()
                tf_hash.update(path.read_bytes())
                result.append(
                    (str(path.relative_to(self._testcase_root)), tf_hash.digest())
                )
        sorted_result = tuple(sorted(result))

        if LOG.getEffectiveLevel() == DEBUG:
            print_hash = sha512()
            print_hash.update(
                repr(sorted_result).encode("utf-8", errors="surrogateescape")
            )
            in_tried = sorted_result in self._tried
            LOG.debug(
                "Testcase hash: %s (%sin cache)",
                print_hash.hexdigest()[:32],
                "" if in_tried else "not ",
            )

        return sorted_result

    def update_tried(self, tried: Iterable[tuple[tuple[str, bytes], ...]]) -> None:
        """Update the list of tried testcase/hash sets. Testcases are hashed with
        SHA-512 and digested to bytes (`hashlib.sha512(testcase).digest()`)

        Arguments:
            tried: Collection of already tried testcase hashes.

        Returns:
            None
        """
        self._tried.update(frozenset(tried))

    def get_tried(self) -> frozenset[tuple[tuple[str, bytes], ...]]:
        """Return the set of tried testcase hashes. Testcases are hashed with SHA-512
        and digested to bytes (`hashlib.sha512(testcase).digest()`)

        Returns:
            Testcase hashes.
        """
        return frozenset(self._tried)

    def dump_testcases(
        self, testcases: list[TestCase], recreate_tcroot: bool = False
    ) -> None:
        """Dump a testcase list to the testcase root on disk.

        Arguments:
            testcases: Testcases to dump.
            recreate_tcroot: if True, delete and recreate tcroot before dumping it.

        Returns:
            None
        """
        if recreate_tcroot:
            rmtree(self._testcase_root)
            self._testcase_root.mkdir()
        for idx, testcase in enumerate(testcases):
            LOG.debug("Extracting testcase %d/%d", idx + 1, len(testcases))
            # NOTE: naming determines load order
            testcase.dump(self._testcase_root / f"{idx:03d}", include_details=True)

    @abstractmethod
    def __iter__(self) -> Iterator[list[TestCase]]:
        """Iterate over potential reductions of testcases according to this strategy.

        The caller should evaluate each reduction yielded, and call `update` with the
        result. The caller owns the testcases yielded, and should call `cleanup` for
        each.

        Yields:
            Testcases with reduction applied.
        """

    @abstractmethod
    def update(self, success: bool) -> None:
        """Inform the strategy whether or not the last reduction yielded was good.

        Arguments:
            success: Whether or not the last reduction was acceptable.

        Returns:
            None
        """

    def __enter__(self) -> Strategy:
        """Enter a runtime context that will automatically call `cleanup` on exit.

        Returns:
            self
        """
        return self

    def __exit__(self, *exc: object) -> None:
        """Exit the runtime context. `cleanup` is called.

        Arguments:

        Returns:
            None
        """
        self.cleanup()

    def cleanup(self) -> None:
        """Destroy all resources held by the strategy.

        Returns:
            None
        """
        rmtree(self._testcase_root)


def _load_strategies() -> dict[str, type[Strategy]]:
    """STRATEGIES is created at the end of this file.

    Returns:
        A mapping of strategy names to strategy class.
    """
    strategies: dict[str, type[Strategy]] = {}
    for entry_point in iter_entry_points("grizzly_reduce_strategies"):
        try:
            strategy_cls = cast(Type[Strategy], entry_point.load())
            assert (
                strategy_cls.name == entry_point.name
            ), f"entry_point name mismatch, check setup.py and {strategy_cls.__name__}"
        except Exception as exc:  # pylint: disable=broad-except
            LOG.debug("error loading strategy type %s: %s", entry_point.name, exc)
            continue
        strategies[entry_point.name] = strategy_cls
    for strategy in DEFAULT_STRATEGIES:
        assert strategy in strategies, (
            f"Unknown entry in DEFAULT_STRATEGIES: {strategy} "
            f"(STRATEGIES: [{','.join(strategies)}])"
        )
    return strategies


def _contains_dd(path: Path) -> bool:
    data = path.read_bytes()
    return b"DDBEGIN" in data and b"DDEND" in data


STRATEGIES = _load_strategies()



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# pylint: disable=protected-access
"""Unit tests for `grizzly.reduce.strategies.beautify`."""
from logging import getLogger

import pytest

from ..common.storage import TestCase
from .strategies.beautify import CSSBeautify, JSBeautify

LOG = getLogger(__name__)


def _test_beautify(cls, interesting, test_name, test_data, reduced, mocker, tmp_path):
    mocker.patch("grizzly.reduce.strategies.beautify._contains_dd", return_value=True)

    with TestCase(test_name, "test-adapter") as test:
        test.add_from_bytes(test_data.encode("ascii"), test.entry_point)
        test.dump(tmp_path / "src", include_details=True)
    best_tests = [TestCase.load(tmp_path / "src")]

    def _interesting(testcases):
        for test in testcases:
            contents = test[test_name].read_bytes().decode("ascii")
            if interesting(contents):
                return True
        return False

    try:
        with cls(best_tests) as sgy:
            for tests in sgy:
                result = _interesting(tests)
                sgy.update(result)
                if result:
                    for test in best_tests:
                        test.cleanup()
                    best_tests = [x.clone() for x in tests]
        assert len(best_tests) == 1
        contents = best_tests[0][test_name].read_bytes().decode("ascii")
        assert contents == reduced
    finally:
        for test in best_tests:
            test.cleanup()


@pytest.mark.parametrize(
    "test_data, reduced",
    [
        pytest.param(
            "try{'a';'R'}catch(e){}\n",
            "try {\n    'a';\n    'R'\n} catch (e) {}\n",
            id="#0: test beautify a .js file",
        ),
        pytest.param(
            "try{\n//DDBEGIN\n'a';'R'\n//DDEND\n}catch(e){}\n",
            "try{\n//DDBEGIN\n'a';\n'R'\n//DDEND\n}catch(e){}\n",
            id="#1: test DDBEGIN/END respected in .js file",
        ),
        pytest.param(
            "try{\r\n'a';'R'}catch(e){}\n",
            "try {\n    'a';\n    'R'\n} catch (e) {}\n",
            id="#2: test that mixed crlf/lf gets converted to lf",
        ),
        pytest.param(
            "try{\r'a';'R'}catch(e){}\n",
            "try {\n    'a';\n    'R'\n} catch (e) {}\n",
            id="#3: test that mixed cr/lf gets converted to lf",
        ),
        pytest.param(
            "try{'a\v';'R'}catch(e){}\n",
            "try {\n    'a\v';\n    'R'\n} catch (e) {}\n",
            id="#4: test beautify with a string containing a \v",
        ),
        pytest.param(
            "try{'a';'R'}\n\ncatch(e){}\n",
            "try {\n    'a';\n    'R'\n} catch (e) {}\n",
            id="#5: test that existing newlines are not preserved",
        ),
    ],
)
def test_beautify_js_1(test_data, reduced, mocker, tmp_path):
    _test_beautify(
        JSBeautify, lambda x: "R" in x, "test.js", test_data, reduced, mocker, tmp_path
    )


@pytest.mark.parametrize(
    "test_data, reduced",
    [
        pytest.param(
            "try{'a';'R'+'R'}catch(e){}\n",
            "try{'a';'R'+'R'}catch(e){}\n",
            id="#6: test beautify js where beautification breaks",
        ),
    ],
)
def test_beautify_js_2(test_data, reduced, mocker, tmp_path):
    _test_beautify(
        JSBeautify,
        lambda x: "'R'+'R'" in x,
        "test.js",
        test_data,
        reduced,
        mocker,
        tmp_path,
    )


@pytest.mark.parametrize(
    "test_data, reduced",
    [
        pytest.param(
            "<script>try{'a';'R'}catch(e){}</script>\n",
            "<script>\ntry {\n    'a';\n    'R'\n} catch (e) {}\n</script>\n",
            id="#7: test beautify js embedded in html",
        ),
        pytest.param(
            "<script>try{'a';'R'}catch(e){}\n",
            "<script>\ntry {\n    'a';\n    'R'\n} catch (e) {}\n",
            id="#8: test beautify js embedded in html with no end",
        ),
        pytest.param(
            "<!--DDBEGIN-->\n<script>\ntry{'a';'R'}catch(e){}\n</script><!--DDEND-->\n",
            "<!--DDBEGIN-->\n<script>\ntry {\n    'a';\n    'R'\n"
            "} catch (e) {}\n</script><!--DDEND-->\n",
            id="#9: test DDBEGIN/END respected for js embedded in html, "
            "DD outside <script>",
        ),
        pytest.param(
            "<script>try{\n//DDBEGIN\n'a';'R'\n//DDEND\n}catch(e){}</script>\n",
            "<script>try{\n//DDBEGIN\n'a';\n'R'\n//DDEND\n}catch(e){}</script>\n",
            id="#10: test DDBEGIN/END respected for js embedded in html, "
            "DD inside <script>",
        ),
        pytest.param(
            "<!--DDBEGIN-->\n<script>\ntry{'a';'R'}catch(e){}\n//DDEND\n",
            "<!--DDBEGIN-->\n<script>\ntry {\n    'a';\n    'R'\n"
            "} catch (e) {}\n//DDEND\n",
            id="#11: test DDBEGIN/END respected for js embedded in html, "
            "DD straddle before <script>",
        ),
        pytest.param(
            "<script>\n//DDBEGIN\ntry{'a';'R'}catch(e){}\n//DDEND\n",
            "<script>\n//DDBEGIN\ntry {\n    'a';\n    'R'\n} catch (e) {}\n//DDEND\n",
            id="#12: test DDBEGIN/END respected for js embedded in html, "
            "DD straddle after <script>",
        ),
        pytest.param(
            "<script>\n//DDBEGIN\ntry{'a';'R'}catch(e){}\n</script>\n<!--DDEND-->\n",
            "<script>\n//DDBEGIN\ntry {\n    'a';\n    'R'\n"
            "} catch (e) {}\n</script>\n<!--DDEND-->\n",
            id="#13: test DDBEGIN/END respected for js embedded in html, "
            "DD straddle after </script>",
        ),
        pytest.param(
            "try{'a';'R'}catch(e){}\n",
            "try{'a';'R'}catch(e){}\n",
            id="#14: test beautify js embedded in html (no <script>)",
        ),
    ],
)
def test_beautify_js_3(test_data, reduced, mocker, tmp_path):
    _test_beautify(
        JSBeautify,
        lambda x: "R" in x,
        "test.html",
        test_data,
        reduced,
        mocker,
        tmp_path,
    )


@pytest.mark.parametrize(
    "test_data, reduced",
    [
        pytest.param(
            "<script>try{'a';'R'}catch(e){}</script><script>'a';'Q'</script>\n",
            "<script>\ntry {\n    'a';\n    'R'\n} catch (e) {}\n</script>"
            "<script>\n'a';\n'Q'\n</script>\n",
            id="#15: test beautify multiple js embedded in html",
        ),
    ],
)
def test_beautify_js_4(test_data, reduced, mocker, tmp_path):
    _test_beautify(
        JSBeautify,
        lambda x: "Q" in x and "R" in x,
        "test.html",
        test_data,
        reduced,
        mocker,
        tmp_path,
    )


@pytest.mark.parametrize(
    "test_data, reduced",
    [
        pytest.param(
            "*,#a{a:0;R:1}\n",
            "*,\n#a {\n  a: 0;\n  R: 1\n}\n",
            id="#0: test beautify a .css file",
        ),
        pytest.param(
            "*,\r\n#a{a:0;R:1}\n",
            "*,\n#a {\n  a: 0;\n  R: 1\n}\n",
            id="#1: test that mixed crlf/lf gets converted to lf",
        ),
        pytest.param(
            "*,\r#a{a:0;R:1}\n",
            "*,\n#a {\n  a: 0;\n  R: 1\n}\n",
            id="#2: test that mixed cr/lf gets converted to lf",
        ),
        pytest.param(
            "*,#a{a:0;\n\nR:1}\n",
            "*,\n#a {\n  a: 0;\n  R: 1\n}\n",
            id="#3: test that existing newlines are not preserved",
        ),
    ],
)
def test_beautify_css_1(test_data, reduced, mocker, tmp_path):
    _test_beautify(
        CSSBeautify,
        lambda x: "R" in x,
        "test.css",
        test_data,
        reduced,
        mocker,
        tmp_path,
    )


@pytest.mark.parametrize(
    "test_data, reduced",
    [
        pytest.param(
            "<style>*,#a{a:0;R:1}</style>\n",
            "<style>\n*,\n#a {\n  a: 0;\n  R: 1\n}\n</style>\n",
            id="#4: test beautify css embedded in html",
        ),
        pytest.param(
            "<style>*,#a{a:0;R:1}\n",
            "<style>\n*,\n#a {\n  a: 0;\n  R: 1\n}\n",
            id="#5: test beautify css embedded in html with no end",
        ),
        pytest.param(
            "<style>\n*,\n#a {\n  a: 0;\n  R: 1\n}\n</style>\n",
            "<style>\n*,\n#a {\n  a: 0;\n  R: 1\n}\n</style>\n",
            id="#6: test already beautified css (beautify does nothing)",
        ),
        pytest.param(
            "*,#a{a:0;R:1}\n",
            "*,#a{a:0;R:1}\n",
            id="#7: test beautify css embedded in html (no <style>)",
        ),
    ],
)
def test_beautify_css_2(test_data, reduced, mocker, tmp_path):
    _test_beautify(
        CSSBeautify,
        lambda x: "R" in x,
        "test.html",
        test_data,
        reduced,
        mocker,
        tmp_path,
    )


@pytest.mark.parametrize(
    "beautify, test_data",
    [
        (JSBeautify, "<script>try{'a';'R'}catch(e){}\n"),
        (CSSBeautify, "<style>*,#a{a:0;R:1}\n"),
    ],
)
def test_no_beautify(beautify, test_data, mocker, tmp_path):
    """test that when beautifiers are not available, the strategies have no effect"""
    mocker.patch.object(beautify, "import_available", False)
    _test_beautify(
        beautify,
        lambda x: "R" in x,
        "test.html",
        test_data,
        test_data,
        mocker,
        tmp_path,
    )



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from __future__ import annotations

from logging import getLogger
from typing import Any

try:
    from pkg_resources import iter_entry_points
except ImportError:
    from .utils import iter_entry_points  # type: ignore


__all__ = ("load_plugin", "scan_plugins", "PluginLoadError")


LOG = getLogger(__name__)


class PluginLoadError(Exception):
    """Raised if loading a plug-in fails"""


def load_plugin(name: str, group: str, base_type: type) -> Any:
    """Load a plug-in.

    Args:
        name: Name of entry point to load.
        group: Group containing entry point.
        base_type: Used to validate loaded objects.

    Returns:
        Loaded plug-in object.
    """
    assert isinstance(base_type, type)
    for entry in iter_entry_points(group):
        if entry.name == name:
            plugin = entry.load()
            LOG.debug("loading %r (%s)", name, base_type.__name__)
            break
    else:
        raise PluginLoadError(f"{name!r} not found in {group!r}")
    if not issubclass(plugin, base_type):
        raise PluginLoadError(f"{name!r} doesn't inherit from {base_type.__name__}")
    return plugin


def scan_plugins(group: str) -> list[str]:
    """Scan for installed plug-ins.

    Args:
        group: Entry point group to scan.

    Returns:
        Names of installed entry points.
    """
    found: list[str] = []
    LOG.debug("scanning %r", group)
    for entry in iter_entry_points(group):
        if entry.name in found:
            # not sure if this can even happen
            raise PluginLoadError(f"Duplicate entry {entry.name!r} in {group!r}")
        found.append(entry.name)
    return found


def scan_target_assets() -> dict[str, tuple[str, ...]]:
    """Scan targets and load collection of supported assets (minimal sanity checking).

    Args:
        None

    Returns:
        Name of target and list of supported assets.
    """
    assets: dict[str, tuple[str, ...]] = {}
    for entry in iter_entry_points("grizzly_targets"):
        assets[entry.name] = entry.load().SUPPORTED_ASSETS
    return assets



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# pylint: disable=protected-access
from platform import system

from ffpuppet import BrowserTerminatedError, BrowserTimeoutError, Debugger, Reason
from pytest import mark, raises

from sapphire import CertificateBundle

from .assets import AssetManager
from .puppet_target import PuppetTarget
from .target import Result, TargetLaunchError, TargetLaunchTimeout


def test_puppet_target_01(mocker, tmp_path):
    """test creating a PuppetTarget"""
    fake_ffp = mocker.patch("grizzly.target.puppet_target.FFPuppet", autospec=True)
    fake_ffp.return_value.reason = Reason.CLOSED
    fake_ffp.return_value.log_length.return_value = 562
    fake_file = tmp_path / "fake"
    fake_file.touch()
    with PuppetTarget(fake_file, 300, 25, 5000) as target:
        assert target.asset_mgr
        assert target.closed
        assert target.launch_timeout == 300
        assert target.log_limit == 25
        assert target.memory_limit == 5000
        assert target.check_result(set()) == Result.NONE
        assert not target.https()
        assert target.log_size() == 1124
        fake_ffp.return_value.log_length.assert_any_call("stderr")
        fake_ffp.return_value.log_length.assert_any_call("stdout")
        assert target.monitor is not None
        target.save_logs(tmp_path / "fake_dest")
        assert fake_ffp.return_value.save_logs.call_count == 1
    assert fake_ffp.return_value.clean_up.call_count == 1
    # with extra args
    with PuppetTarget(fake_file, 1, 1, 1, rr=True, fake=1) as target:
        pass


def test_puppet_target_02(mocker, tmp_path):
    """test PuppetTarget.launch()"""
    fake_ffp = mocker.patch("grizzly.target.puppet_target.FFPuppet", autospec=True)
    fake_file = tmp_path / "fake"
    fake_file.touch()
    # test providing prefs.js
    with PuppetTarget(fake_file, 300, 25, 5000) as target:
        # launch success
        target.launch("launch_target_page")
        assert fake_ffp.return_value.launch.call_count == 1
        assert fake_ffp.return_value.close.call_count == 0
        target.close()
        # launch timeout
        fake_ffp.reset_mock()
        fake_ffp.return_value.launch.side_effect = BrowserTimeoutError("timeout")
        with raises(TargetLaunchTimeout, match="timeout"):
            target.launch("launch_target_page")
        assert fake_ffp.return_value.save_logs.call_count == 0
        # launch failure
        fake_ffp.reset_mock()
        (tmp_path / "log_stderr.txt").write_text("fake log")
        (tmp_path / "log_stdout.txt").write_text("fake log")
        mocker.patch(
            "grizzly.target.puppet_target.mkdtemp",
            autospec=True,
            return_value=str(tmp_path),
        )
        fake_ffp.return_value.launch.side_effect = BrowserTerminatedError("fail")
        with raises(TargetLaunchError, match="fail"):
            target.launch("launch_target_page")
        assert fake_ffp.return_value.save_logs.call_count == 1


@mark.parametrize(
    "healthy, reason, ignore, result, closes",
    [
        # running as expected - no failures
        (True, None, [], Result.NONE, 0),
        # browser process closed
        (False, Reason.CLOSED, [], Result.NONE, 1),
        # browser process crashed
        (False, Reason.ALERT, [], Result.FOUND, 1),
        # browser exit with no crash logs
        (False, Reason.EXITED, [], Result.NONE, 1),
        # ffpuppet check failed
        (False, Reason.WORKER, [], Result.FOUND, 1),
        # ffpuppet check ignored (memory)
        (False, Reason.WORKER, ["memory"], Result.IGNORED, 1),
        # ffpuppet check ignored (log-limit)
        (False, Reason.WORKER, ["log-limit"], Result.IGNORED, 1),
    ],
)
def test_puppet_target_03(mocker, tmp_path, healthy, reason, ignore, result, closes):
    """test PuppetTarget.check_result()"""
    fake_ffp = mocker.patch("grizzly.target.puppet_target.FFPuppet", autospec=True)
    fake_file = tmp_path / "fake"
    fake_file.touch()
    if "memory" in ignore:
        fake_ffp.return_value.available_logs.return_value = "ffp_worker_memory_usage"
    elif "log-limit" in ignore:
        fake_ffp.return_value.available_logs.return_value = "ffp_worker_log_size"
    fake_ffp.return_value.is_healthy.return_value = healthy
    fake_ffp.return_value.reason = reason
    with PuppetTarget(fake_file, 300, 25, 5000) as target:
        assert target.check_result(ignore) == result
    assert fake_ffp.return_value.close.call_count == closes


@mark.parametrize(
    "healthy, usage, os_name, killed, debugger",
    [
        # skip sending SIGABRT on unsupported OSs
        (True, [(1234, 90)], "Windows", 0, Debugger.NONE),
        # skip idle check if target is in a bad state
        (False, [], "Linux", 0, Debugger.NONE),
        # send SIGABRT to hung process
        (True, [(234, 10), (236, 75), (238, 60)], "Linux", 1, Debugger.NONE),
        # Don't send SIGABRT when using a debugger
        (True, [(236, 75)], "Linux", 0, Debugger.RR),
        # ignore idle timeout (close don't abort)
        (True, [(234, 10)], "Linux", 0, Debugger.NONE),
    ],
)
def test_puppet_target_04(mocker, tmp_path, healthy, usage, os_name, killed, debugger):
    """test PuppetTarget.handle_hang()"""
    mocker.patch(
        "grizzly.target.puppet_target.system", autospec=True, return_value=os_name
    )
    fake_ffp = mocker.patch("grizzly.target.puppet_target.FFPuppet", autospec=True)
    fake_kill = mocker.patch("grizzly.target.puppet_target.kill", autospec=True)
    # raise OSError for code coverage
    fake_kill.side_effect = OSError
    fake_file = tmp_path / "fake"
    fake_file.touch()
    fake_ffp.return_value.cpu_usage.return_value = usage
    fake_ffp.return_value.is_healthy.return_value = healthy
    with PuppetTarget(fake_file, 300, 25, 5000) as target:
        target._debugger = debugger
        target.handle_hang()
    assert fake_ffp.return_value.is_healthy.call_count == 1
    assert fake_ffp.return_value.close.call_count == 1
    assert fake_ffp.return_value.cpu_usage.call_count == (1 if usage else 0)
    assert fake_kill.call_count == fake_ffp.return_value.wait.call_count == killed


@mark.skipif(system() != "Linux", reason="Linux only")
def test_puppet_target_05(mocker, tmp_path):
    """test PuppetTarget.dump_coverage()"""
    fake_ffp = mocker.patch("grizzly.target.puppet_target.FFPuppet", autospec=True)
    fake_file = tmp_path / "fake"
    fake_file.touch()
    with PuppetTarget(fake_file, 300, 25, 5000) as target:
        target.dump_coverage()
    assert fake_ffp.return_value.dump_coverage.call_count == 1


def test_puppet_target_06(mocker, tmp_path):
    """test PuppetTarget.monitor"""
    fake_ffp = mocker.patch("grizzly.target.puppet_target.FFPuppet", autospec=True)
    fake_file = tmp_path / "fake"
    fake_file.touch()
    with PuppetTarget(fake_file, 300, 25, 5000) as target:
        fake_ffp.return_value.is_running.return_value = False
        fake_ffp.return_value.is_healthy.return_value = False
        assert target.monitor is not None
        assert not target.monitor.is_healthy()
        assert not target.monitor.is_running()
        fake_ffp.return_value.is_running.return_value = True
        fake_ffp.return_value.is_healthy.return_value = True
        assert target.monitor.is_healthy()
        assert target.monitor.is_running()
        fake_ffp.return_value.launches = 123
        assert target.monitor.launches == 123
        fake_ffp.return_value.log_length.return_value = 100
        assert target.monitor.log_length("stdout") == 100
        target.monitor.clone_log("somelog")
        assert fake_ffp.return_value.clone_log.call_count == 1


def test_puppet_target_07(mocker, tmp_path):
    """test PuppetTarget.monitor.is_idle()"""
    fake_ffp = mocker.patch("grizzly.target.puppet_target.FFPuppet", autospec=True)
    fake_ffp.return_value.cpu_usage.return_value = [(999, 30), (998, 20), (997, 10)]
    fake_file = tmp_path / "fake"
    fake_file.touch()
    with PuppetTarget(fake_file, 300, 25, 5000) as target:
        assert not target.monitor.is_idle(0)
        assert not target.monitor.is_idle(25)
        assert target.monitor.is_idle(50)


def test_puppet_target_08(mocker, tmp_path):
    """test PuppetTarget.process_assets()"""
    mocker.patch("grizzly.target.puppet_target.FFPuppet", autospec=True)
    fake_file = tmp_path / "fake"
    fake_file.write_text("1\n2\n")
    # no prefs file provided
    with PuppetTarget(fake_file, 300, 25, 5000) as target:
        assert target.asset_mgr.get("prefs") is None
        target.process_assets()
        asset = target.asset_mgr.get("prefs")
        assert asset
        assert asset.is_file()
        assert asset.name == "prefs.js"
    # prefs file provided
    with AssetManager(base_path=tmp_path) as asset_mgr:
        asset_mgr.add("prefs", fake_file)
        with PuppetTarget(fake_file, 300, 25, 5000) as target:
            target.asset_mgr = asset_mgr
            target.process_assets()
            asset = target.asset_mgr.get("prefs")
            assert asset
            assert asset.name == "fake"
    # abort tokens file provided
    with AssetManager(base_path=tmp_path) as asset_mgr:
        asset_mgr.add("abort-tokens", fake_file)
        with PuppetTarget(fake_file, 300, 25, 5000) as target:
            # ignore E1101: (pylint 2.9.3 bug?)
            #    Method 'add_abort_token' has no 'call_count' member (no-member)
            # pylint: disable=no-member
            assert target._puppet.add_abort_token.call_count == 0
            target.asset_mgr = asset_mgr
            target.process_assets()
            asset = target.asset_mgr.get("abort-tokens")
            assert asset
            assert asset.is_file()
            assert asset.name == "fake"
            assert target._puppet.add_abort_token.call_count == 2


@mark.parametrize(
    "pernosco, rr, valgrind",
    [
        # No debugger selected
        (False, False, False),
        # Pernosco selected
        (True, False, False),
        # rr selected
        (False, True, False),
        # Valgrind selected
        (False, False, True),
    ],  # pylint: disable=invalid-name
)
def test_puppet_target_09(
    mocker, tmp_path, pernosco, rr, valgrind
):  # pylint: disable=invalid-name
    """test PuppetTarget debugger args"""
    fake_ffp = mocker.patch("grizzly.target.puppet_target.FFPuppet", autospec=True)
    fake_file = tmp_path / "fake"
    fake_file.touch()
    with PuppetTarget(
        fake_file, 30, 25, 500, pernosco=pernosco, rr=rr, valgrind=valgrind
    ) as _:
        pass
    if pernosco:
        assert fake_ffp.call_args[-1]["debugger"] == Debugger.PERNOSCO
    elif rr:
        assert fake_ffp.call_args[-1]["debugger"] == Debugger.RR
    elif valgrind:
        assert fake_ffp.call_args[-1]["debugger"] == Debugger.VALGRIND
    else:
        assert fake_ffp.call_args[-1]["debugger"] == Debugger.NONE


@mark.parametrize(
    "asset, env",
    [
        # suppressions via asset
        (True, False),
        # suppressions via env
        (False, True),
        # suppressions via both asset and env (asset should be preferred)
        (True, True),
        # missing suppressions file
        (False, False),
    ],
)
def test_puppet_target_10(tmp_path, asset, env):
    """test PuppetTarget.process_assets() - configure sanitizer suppressions"""
    fake_file = tmp_path / "fake"
    fake_file.touch()
    supp_asset = tmp_path / "supp_asset"
    supp_env = tmp_path / "supp_env"
    with AssetManager(base_path=tmp_path) as asset_mgr:
        asset_mgr.add("prefs", fake_file)
        if asset:
            supp_asset.touch()
            asset_mgr.add("lsan-suppressions", supp_asset)
        with PuppetTarget(fake_file, 300, 25, 5000) as target:
            target.environ["TSAN_OPTIONS"] = "a=1"
            if env:
                supp_env.touch()
                target.environ["LSAN_OPTIONS"] = f"suppressions='{supp_env}'"
            else:
                target.environ["LSAN_OPTIONS"] = "suppressions='missing'"
            target.asset_mgr = asset_mgr
            target.process_assets()
            if asset:
                assert (
                    f"suppressions='{target.asset_mgr.path / supp_asset.name}'"
                    in target.environ["LSAN_OPTIONS"]
                )
            elif env:
                assert (
                    f"suppressions='{target.asset_mgr.path / supp_env.name}'"
                    in target.environ["LSAN_OPTIONS"]
                )
            else:
                assert not asset_mgr.get("lsan-suppressions")


def test_puppet_target_11(tmp_path):
    """test PuppetTarget.filtered_environ()"""
    fake_file = tmp_path / "fake"
    fake_file.touch()
    with PuppetTarget(fake_file, 300, 25, 5000) as target:
        target.environ = {
            "TRACKED": "1",
            "ASAN_OPTIONS": "external_symbolizer_path='a':no_remove='b'",
            "LSAN_OPTIONS": "suppressions='a'",
            "EMPTY": "",
        }
        assert "EMPTY" not in target.filtered_environ()
        assert "TRACKED" in target.filtered_environ()
        assert "ASAN_OPTIONS" in target.filtered_environ()
        assert "LSAN_OPTIONS" not in target.filtered_environ()


@mark.parametrize(
    "base, extra, result",
    [
        # empty
        ({}, {}, {}),
        # empty extra
        ({"a": "1"}, {}, {"a": "1"}),
        # empty base
        ({}, {"a": "1"}, {"a": "1"}),
        # merge
        ({"a": "1"}, {"b": "2"}, {"a": "1", "b": "2"}),
        # name collision, favor base
        ({"a": "1"}, {"a": "2"}, {"a": "1"}),
        # name collision and merge
        ({"a": "1"}, {"a": "2", "b": "2"}, {"a": "1", "b": "2"}),
    ],
)
def test_puppet_target_12(tmp_path, base, extra, result):
    """test PuppetTarget.merge_environment()"""
    fake_file = tmp_path / "fake"
    fake_file.touch()
    with PuppetTarget(fake_file, 300, 25, 5000) as target:
        target.environ = base
        target.merge_environment(extra)
        assert target.environ == result


@mark.parametrize(
    "base, extra, result",
    [
        # collision with existing option
        ({"ASAN_OPTIONS": "a=1"}, {"ASAN_OPTIONS": "a=2"}, ["a=1"]),
        # add option from extra
        ({"ASAN_OPTIONS": "a=1"}, {"ASAN_OPTIONS": "b=2:c3"}, ["a=1", "b=2", "c=3"]),
        # add option from extra
        ({"ASAN_OPTIONS": "a=1:c=3"}, {"ASAN_OPTIONS": "b=2"}, ["a=1", "b=2", "c=3"]),
    ],
)
def test_puppet_target_13(tmp_path, base, extra, result):
    """test PuppetTarget.merge_environment() - merge sanitizer options"""
    fake_file = tmp_path / "fake"
    fake_file.touch()
    with PuppetTarget(fake_file, 300, 25, 5000) as target:
        target.environ = base
        target.merge_environment(extra)
        for opt in target.environ["ASAN_OPTIONS"].split(":"):
            assert opt in result


def test_puppet_target_14(mocker, tmp_path):
    """test PuppetTarget.dump_coverage() - skip on unsupported platform"""
    mocker.patch("grizzly.target.puppet_target.system", return_value="foo")
    fake_file = tmp_path / "fake"
    fake_file.touch()
    with PuppetTarget(fake_file, 300, 25, 5000) as target:
        target.dump_coverage()


@mark.parametrize(
    "certutil, certs",
    [
        # certutil and cert bundle available
        (True, True),
        # missing certutil
        (False, True),
        # no cert bundle
        (True, False),
    ],
)
def test_puppet_target_15(mocker, tmp_path, certutil, certs):
    """test PuppetTarget - HTTPS support"""
    mocker.patch(
        "grizzly.target.puppet_target.certutil_available", return_value=certutil
    )

    fake_file = tmp_path / "fake"
    fake_file.touch()

    certs_bundle = mocker.Mock(spec_set=CertificateBundle) if certs else None
    with PuppetTarget(fake_file, 300, 25, 5000, certs=certs_bundle) as target:
        assert target.https() == (certutil and certs)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""Grizzly Reduction exceptions."""
from ..common.utils import Exit


class GrizzlyReduceBaseException(Exception):
    """Base for other Grizzly Reducer specific exceptions."""

    def __init__(self, msg: str, code: Exit = Exit.ERROR) -> None:
        super().__init__()
        self.msg = msg
        self.code = code


class NotReproducible(GrizzlyReduceBaseException):
    """Crash was not observed when expected during reduction."""

    def __init__(self, msg: str) -> None:
        super().__init__(msg, code=Exit.FAILURE)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""CLI argument parsing for Grizzly reduction.
"""
from argparse import Namespace
from logging import getLogger
from pathlib import Path

from ..common.reporter import Quality
from ..replay.args import LOCAL_INPUT_HELP, ReplayCommonArgs
from .strategies import DEFAULT_STRATEGIES, STRATEGIES

LOG = getLogger(__name__)


class ReduceCommonArgs(ReplayCommonArgs):
    """Argument parser for `grizzly.reduce`.

    Takes all arguments defined for `grizzly.replay`, and a few specific to reduction.
    """

    def __init__(self) -> None:
        """Initialize argument parser."""
        super().__init__()

        # these arguments have defaults that vary from ReplayCommonArgs
        self.parser.set_defaults(output=Path.cwd())

        reduce_args = self.parser.add_argument_group("Reduce Arguments")
        reduce_args.add_argument(
            "--no-analysis",
            action="store_true",
            help="Disable analysis to auto-set --repeat/--min-crashes.",
        )
        reduce_args.add_argument(
            "--report-period",
            type=int,
            help="Periodically report the best testcase for long-running strategies."
            " (value in seconds, default: no)",
        )
        reduce_args.add_argument(
            "--static-timeout",
            action="store_true",
            help="Disable automatically updating the iteration timeout.",
        )
        reduce_args.add_argument(
            "--strategy",
            nargs="+",
            choices=STRATEGIES,
            default=DEFAULT_STRATEGIES,
            dest="strategies",
            help="One or more strategies. (default: %(default)s)",
        )

    def sanity_check(self, args: Namespace) -> None:
        """Sanity check reducer args.

        Arguments:
            args: Result from `parse_args()`.

        Raises:
            SystemExit: on error, `ArgumentParser.error()` is called, which will exit.
        """
        super().sanity_check(args)

        if args.report_period is not None:
            if args.report_period <= 0:
                self.parser.error("Invalid --report-period (value is in seconds)")
            if args.report_period < 60:
                self.parser.error("Very short --report-period (value is in seconds)")

        if not args.no_analysis:
            # analysis is enabled, but repeat/min_crashes specified. doesn't make sense
            errors = []
            if args.repeat != self.parser.get_default("repeat"):
                errors.append("'--repeat'")
            if args.min_crashes != self.parser.get_default("min_crashes"):
                errors.append("'--min-crashes'")
            if errors:
                error_str = " and ".join(errors)
                LOG.warning(
                    "%s specified, with analysis enabled, they will be ignored",
                    error_str,
                )


class ReduceArgs(ReduceCommonArgs):
    # NOTE: If updated changes may also need to be added to ReplayArgs
    def __init__(self) -> None:
        super().__init__()
        self.parser.add_argument("input", type=Path, nargs="+", help=LOCAL_INPUT_HELP)

        self.launcher_grp.add_argument(
            "--entry-point",
            type=Path,
            help="Specify file to use as testcase entry-point when it cannot be"
            " automatically determined.",
        )

    def sanity_check(self, args: Namespace) -> None:
        super().sanity_check(args)

        for test in args.input:
            if not test.exists():
                self.parser.error(f"'{test}' does not exist")

        if args.no_harness and len(args.input) > 1:
            self.parser.error("'--no-harness' cannot be used with multiple testcases")


class ReduceFuzzManagerIDArgs(ReduceCommonArgs):
    def __init__(self) -> None:
        """Initialize argument parser."""
        super().__init__()
        self.parser.add_argument("input", type=int, help="FuzzManager ID to reduce")

        self.parser.add_argument(
            "--no-repro-quality",
            choices=[x.value for x in Quality],
            default=Quality.NOT_REPRODUCIBLE.value,
            type=int,
            help="Quality value reported when issue does not reproduce "
            "(default: %(default)s).",
        )

        self.parser.add_argument("--test-ext", help="Overwrite testcase file extension")

        self.parser.add_argument(
            "--test-index",
            default=[],
            type=int,
            nargs="+",
            help="Select a testcase to run when multiple testcases are loaded. "
            "Testscases are ordered oldest to newest. "
            "0 == oldest, n-1 == most recent (default: run all testcases)",
        )

    def sanity_check(self, args: Namespace) -> None:
        super().sanity_check(args)

        if args.no_harness and len(args.test_index) > 1:
            self.parser.error(
                "'--test-index' only supports a single value with '--no-harness'"
            )


class ReduceFuzzManagerIDQualityArgs(ReduceFuzzManagerIDArgs):
    def __init__(self) -> None:
        """Initialize argument parser."""
        super().__init__()
        self.parser.add_argument(
            "--quality",
            choices=[x.value for x in Quality],
            type=int,
            help="Only try crashes with a given quality value.",
        )



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from .assets import AssetManager
from .target import Target


class SimpleTarget(Target):
    TRACKED_ENVVARS = ("TEST_INC",)

    def _cleanup(self):
        pass

    def check_result(self, ignored):
        pass

    def close(self, force_close=False):
        pass

    @property
    def closed(self):
        return True

    def create_report(self, is_hang=False, unstable=False):
        pass

    def dump_coverage(self, timeout=0):
        pass

    def handle_hang(self, ignore_idle=True, ignore_timeout=False):
        return False

    def https(self):
        return self._https

    def launch(self, location):
        pass

    def log_size(self):
        return 0

    @property
    def monitor(self):
        return self._monitor

    def merge_environment(self, extra):
        pass

    def process_assets(self):
        pass

    def save_logs(self, dst):
        pass


def test_target_01(tmp_path):
    """test creating a simple Target"""
    fake_file = tmp_path / "fake"
    fake_file.touch()
    with SimpleTarget(fake_file, 10, 2, 3) as target:
        assert target.binary == fake_file
        assert target.asset_mgr
        org_path = target.asset_mgr.path
        target.asset_mgr = AssetManager(base_path=tmp_path)
        assert target.asset_mgr.path != org_path
        assert not target.environ
        assert not target.filtered_environ()
        assert target.launch_timeout == 10
        assert target.log_size() == 0
        assert target.log_limit == 2
        assert target.memory_limit == 3
        assert target.monitor is None
        # test stubs
        target.reverse(1, 2)


def test_target_02(mocker, tmp_path):
    """test loading TRACKED_ENVVARS"""
    mocker.patch.dict("grizzly.target.target.environ", {"SKIP": "x", "TEST_INC": "1"})
    fake_file = tmp_path / "fake"
    fake_file.touch()
    with SimpleTarget(fake_file, 321, 2, 3) as target:
        assert target.environ
        assert "SKIP" not in target.environ
        assert target.environ["TEST_INC"] == "1"
        filtered = target.filtered_environ()
        assert filtered
        assert "SKIP" not in filtered
        assert filtered["TEST_INC"] == "1"


def test_target_03():
    """test Target.scan_environment()"""
    assert not Target.scan_environment({"a": "1"}, ())
    assert not Target.scan_environment({}, ("a",))
    assert Target.scan_environment({"a": "1", "b": "2"}, ("a",)) == {"a": "1"}



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from __future__ import annotations

from logging import getLogger
from pathlib import Path
from shutil import copyfile, copytree, move, rmtree
from tempfile import mkdtemp

__all__ = ("AssetError", "AssetManager")
__author__ = "Tyson Smith"
__credits__ = ["Tyson Smith"]

LOG = getLogger(__name__)


class AssetError(Exception):
    """Raised by AssetManager"""


class AssetManager:
    __slots__ = ("assets", "path")

    def __init__(self, base_path: Path | None = None) -> None:
        self.assets: dict[str, str] = {}
        self.path = Path(mkdtemp(prefix="assets_", dir=base_path))

    def __enter__(self) -> AssetManager:
        return self

    def __exit__(self, *exc: object) -> None:
        self.cleanup()

    def add(self, asset: str, path: Path, copy: bool = True) -> Path:
        """Add asset to the AssetManager.

        Args:
            asset: Name of asset.
            path: Location on disk.
            copy: Copy or move the content.

        Returns:
            Path to the asset on the filesystem.
        """
        assert asset
        if not path.exists():
            raise OSError(f"'{path}' does not exist")
        dst = self.path / path.name
        # remove existing asset with the same name
        if asset in self.assets:
            LOG.debug("asset %r exists, removing existing", asset)
            self.remove(asset)
        # avoid overwriting data that is part of an existing asset
        if dst.exists():
            raise AssetError(f"{asset}: '{path.name}' already exists")
        if copy:
            if path.is_file():
                copyfile(path, dst)
            else:
                copytree(path, dst)
        else:
            # Python 3.9+: move() accepts a path-like object for both src and dst
            move(str(path.resolve()), str(self.path.resolve()))
        self.assets[asset] = path.name
        LOG.debug("%s asset %r to '%s'", "copied" if copy else "moved", asset, dst)
        return dst

    def add_batch(self, assets: list[list[str]]) -> None:
        """Add collection of assets to the AssetManager.

        Args:
            assets: List of list that contain asset, path pairs.

        Returns:
            None
        """
        for asset, path in assets:
            self.add(asset, Path(path))

    def cleanup(self) -> None:
        """Remove asset files from filesystem.

        Args:
            None

        Returns:
            None
        """
        if self.path:
            rmtree(self.path, ignore_errors=True)
            self.assets.clear()

    def get(self, asset: str) -> Path | None:
        """Get path to content on filesystem for given asset.

        Args:
            asset: Asset to lookup.

        Returns:
            Path to asset content or None if asset does not exist.
        """
        item = self.assets.get(asset, None)
        return self.path / item if item else None

    def is_empty(self) -> bool:
        """Check if AssetManager contains entries.

        Args:
            None

        Returns:
            True if AssetManager contains entries else False.
        """
        return not self.assets

    @classmethod
    def load(
        cls, assets: dict[str, str], src_path: Path, base_path: Path | None = None
    ) -> AssetManager:
        """Load assets from filesystem.

        Args:
            asset: Asset paths on filesystem relative to src_path, keyed on asset name.
            src_path: Path to scan for assets.
            base_path: Base path to use to create local storage.

        Returns:
            AssetManager populated with contents provided by assets argument.
        """
        obj = cls(base_path=base_path)
        for asset, src_name in assets.items():
            obj.add(asset, src_path / src_name)
        return obj

    def remove(self, asset: str) -> None:
        """Remove asset from AssetManager if asset exists.

        Args:
            asset: Asset to remove.

        Returns:
            None
        """
        local_path = self.assets.pop(asset, None)
        if local_path:
            path = self.path / local_path
            if path.is_file():
                path.unlink()
            else:
                rmtree(path, ignore_errors=True)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""
Loki fuzzing library
"""
from __future__ import annotations

from logging import ERROR, INFO, basicConfig, getLogger
from os import SEEK_END
from pathlib import Path
from random import choice, getrandbits, randint, sample
from shutil import copy
from struct import pack, unpack
from tempfile import SpooledTemporaryFile, mkdtemp
from time import perf_counter, strftime
from typing import IO, TYPE_CHECKING

if TYPE_CHECKING:
    from argparse import Namespace

__author__ = "Tyson Smith"

LOG = getLogger(__name__)


class Loki:
    BYTE_ORDERS = frozenset(("<", ">", "@", "!", "="))

    __slots__ = ("aggr", "byte_order")

    def __init__(self, aggression: float = 0.0, byte_order: str | None = None) -> None:
        """
        Arguments:
            aggression: Amount of fuzzing to perform. 0 for 0% up to 1.0 for 100%.
            byte_order: Used to indicate big or little endian when mutating multiple
                        bytes at once.
        """
        assert 0 <= aggression <= 1
        assert byte_order is None or byte_order in self.BYTE_ORDERS
        self.aggr = aggression
        self.byte_order = byte_order

    @staticmethod
    def _fuzz_data(in_data: bytes, byte_order: str) -> bytes:
        """Fuzz data.

        Args:
            in_data: Data to fuzz.
            byte_order: Byte order to use.

        Returns:
            Fuzzed data.
        """
        data_size = len(in_data)
        if data_size == 1:
            pack_unit = "B"
            mask = 0xFF
        elif data_size == 2:
            pack_unit = f"{byte_order}H"
            mask = 0xFFFF
        elif data_size == 4:
            pack_unit = f"{byte_order}I"
            mask = 0xFFFFFFFF
        else:
            raise RuntimeError(f"Unsupported data size: {data_size}")

        fuzz_op = randint(0, 4)
        if fuzz_op == 0:  # boundary
            out_data: int = (2 ** randint(2, data_size * 8)) + randint(-2, 2)
        elif fuzz_op == 1:  # arithmetic
            out_data = unpack(pack_unit, in_data)[0] + randint(-10, 10)
        elif fuzz_op == 2:  # interesting byte, short or int
            out_data = choice((0, 1, mask // 2, (mask // 2) + 1, mask))
        elif fuzz_op == 3:  # random byte, short or int
            out_data = getrandbits(32)
        elif fuzz_op == 4:
            if data_size == 1:  # toggle
                out_data = unpack(pack_unit, in_data)[0] ^ (2 ** getrandbits(3))
            elif data_size == 2:  # multiple of 8
                out_data = randint(1, 0x1FFF) * 8
            elif data_size == 4:  # multiple of 8
                out_data = randint(1, 0x1FFFFFFF) * 8
            else:  # pragma: no cover
                # this should be unreachable
                raise RuntimeError(f"Unsupported data size: {data_size}")
        else:  # pragma: no cover
            # this should be unreachable
            raise AssertionError(f"Invalid fuzz op: {fuzz_op}")

        return pack(pack_unit, out_data & mask)

    def _fuzz(self, tgt_fp: IO[bytes]) -> None:
        """Fuzz file data.

        Args:
            tgt_fp: Open file object.

        Returns:
            None
        """
        tgt_fp.seek(0, SEEK_END)
        length = tgt_fp.tell()
        if length < 1:
            LOG.debug("cannot fuzz empty file")
            return
        # minimum number of mutations should be 1
        max_mutations = max(round(length * self.aggr), 1)
        mutations = randint(1, max_mutations)
        LOG.debug(
            "%d of a possible %d mutations will be performed", mutations, max_mutations
        )
        if self.byte_order is not None:
            assert self.byte_order in self.BYTE_ORDERS
            byte_order = self.byte_order
        else:
            LOG.debug("using random byte order")
            byte_order = "<" if getrandbits(1) else ">"
        for count, idx in enumerate(sample(range(length), k=mutations)):
            # every few iterations randomly attempt multi-byte mutations
            if count % 10 == 0 and getrandbits(1):
                max_size = length - idx
                # target multiple bytes if possible
                if max_size > 3 and getrandbits(1):
                    size = 4
                elif max_size > 1:
                    size = 2
                else:
                    size = 1
            else:
                # target single byte
                size = 1
            # perform mutation
            tgt_fp.seek(idx)
            out_data = self._fuzz_data(tgt_fp.read(size), byte_order)
            tgt_fp.seek(idx)
            tgt_fp.write(out_data)

    def fuzz_data(self, data: bytes) -> bytes:
        """Create a fuzzed copy of the provided data.

        Args:
            data: Data to be fuzzed.

        Returns:
            Fuzzed data.
        """
        assert isinstance(data, bytes)
        # open a temp file in memory for fuzzing
        with SpooledTemporaryFile(max_size=0x800000, mode="r+b") as tmp_fp:
            tmp_fp.write(data)
            self._fuzz(tmp_fp)
            tmp_fp.seek(0)
            return tmp_fp.read()

    def fuzz_file(
        self, src: Path, count: int, dst: Path, ext: str | None = None
    ) -> bool:
        """Create fuzzed copies of the provided file.

        Args:
            src: Template file containing data to be fuzzed.
            count: Number of fuzzed copies to create.
            dst: Directory to store output.
            ext: File extension to use.

        Returns:
            True if successful otherwise False.
        """
        try:
            if src.stat().st_size < 1:
                LOG.error("Input must be at least 1 byte long")
                return False
        except FileNotFoundError:
            LOG.error("'%s' does not exist!", src)
            return False
        if ext is None:
            ext = src.suffix
        for i in range(count):
            out_file = dst / f"{i:06d}_fuzzed{ext}"
            copy(src, out_file)
            with out_file.open("r+b") as out_fp:
                self._fuzz(out_fp)
        return True

    @classmethod
    def main(cls, args: Namespace) -> int:
        basicConfig(format="", level=INFO if not args.quiet else ERROR)
        LOG.info("Starting Loki @ %s", strftime("%Y-%m-%d %H:%M:%S"))
        LOG.info("Target template is '%s'", args.input.resolve())
        out_dir = args.output
        if out_dir is None:
            out_dir = Path(mkdtemp(prefix=strftime("loki_%m-%d_%H-%M_"), dir="."))
        out_dir.mkdir(exist_ok=True)
        LOG.info("Output directory is '%s'", out_dir.resolve())
        count = max(args.count, 1)
        LOG.info("Generating %d fuzzed test cases...", count)
        loki = Loki(aggression=args.aggression, byte_order=args.byte_order)
        try:
            start_time = perf_counter()
            success = loki.fuzz_file(args.input, count, out_dir)
            finish_time = perf_counter() - start_time
            LOG.info("Done. Total run time %gs", finish_time)
            if success:
                LOG.info("About %gs per file", finish_time / count)
        except KeyboardInterrupt:  # pragma: no cover
            LOG.info("Ctrl+C detected.")
            success = False
        return 0 if success else 1



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from __future__ import annotations

from enum import IntEnum, unique
from logging import getLogger
from time import time
from typing import TYPE_CHECKING

from .common.iomanager import IOManager
from .common.runner import Runner
from .common.status import STATUS_DB_FUZZ, Status
from .target import Result, Target

if TYPE_CHECKING:
    from sapphire import Sapphire

    from .adapter import Adapter
    from .common.reporter import Reporter
    from .common.storage import TestCase

__all__ = ("SessionError", "LogOutputLimiter", "Session")
__author__ = "Tyson Smith"
__credits__ = ["Tyson Smith", "Jesse Schwartzentruber"]


LOG = getLogger(__name__)


class SessionError(Exception):
    """The base class for exceptions raised by Session"""


@unique
class LogRate(IntEnum):
    # quickly reduce the amount of output
    NORMAL = 0
    # display status every iteration
    VERBOSE = 1


class LogOutputLimiter:
    __slots__ = (
        "_delay",
        "_iterations",
        "_launches",
        "_multiplier",
        "_rate",
        "_time",
    )

    def __init__(
        self,
        delay: int = 300,
        delta_multiplier: int = 2,
        rate: LogRate = LogRate.NORMAL,
    ) -> None:
        self._delay = delay  # maximum time delay between output
        self._iterations = 1  # next iteration to trigger output
        self._launches = 1  # next launch to trigger output
        self._multiplier = delta_multiplier  # rate to decrease output (iterations)
        self._rate = rate
        self._time = time()

    def ready(self, cur_iter: int, launches: int) -> bool:
        # calculate if a status line should be output
        if self._rate == LogRate.VERBOSE:
            return True
        ready = False
        if cur_iter >= self._iterations:
            ready = True
            self._iterations *= self._multiplier
        elif launches >= self._launches or time() - self._delay >= self._time:
            ready = True
        if ready:
            self._time = time()
            self._launches = launches + 1
        return ready


class Session:
    # display warning when target log files exceed limit (25MB)
    TARGET_LOG_SIZE_WARN = 0x1900000

    __slots__ = (
        "_coverage",
        "_relaunch",
        "_report_size",
        "adapter",
        "iomanager",
        "reporter",
        "server",
        "status",
        "target",
    )

    def __init__(
        self,
        adapter: Adapter,
        reporter: Reporter,
        server: Sapphire,
        target: Target,
        coverage: bool = False,
        enable_profiling: bool = False,
        relaunch: int = 1,
        report_limit: int = 0,
        report_size: int = 1,
    ) -> None:
        assert relaunch > 0
        assert report_limit >= 0
        assert report_size > 0
        self._coverage = coverage
        self._relaunch = relaunch
        self.adapter = adapter
        self.iomanager = IOManager(report_size=report_size)
        self.reporter = reporter
        self.server = server
        self.status = Status.start(
            STATUS_DB_FUZZ,
            enable_profiling=enable_profiling,
            report_limit=report_limit,
        )
        self.target = target

    def __enter__(self) -> Session:
        return self

    def __exit__(self, *exc: object) -> None:
        self.close()

    def close(self) -> None:
        # perform final status report
        self.status.report(force=True)
        self.iomanager.cleanup()

    def display_status(self, log_limiter: LogOutputLimiter) -> None:
        if self.adapter.remaining is not None:
            LOG.info(
                "[I%04d-L%02d-R%02d] %s",
                self.status.iteration,
                self.adapter.remaining,
                self.status.results.total,
                self.status.test_name,
            )
        elif log_limiter.ready(self.status.iteration, self.target.monitor.launches):
            LOG.info("I%04d-R%02d ", self.status.iteration, self.status.results.total)

    def generate_testcase(self) -> TestCase:
        LOG.debug("calling iomanager.create_testcase()")
        test = self.iomanager.create_testcase(self.adapter.name)
        LOG.debug("calling adapter.generate()")
        with self.status.measure("generate"):
            self.adapter.generate(test, self.iomanager.server_map)
        self.status.test_name = test.input_fname
        return test

    def run(
        self,
        ignore: set[str],
        time_limit: int,
        input_path: str | None = None,
        iteration_limit: int = 0,
        no_harness: bool = False,
        result_limit: int = 0,
        runtime_limit: int = 0,
        log_rate: LogRate = LogRate.NORMAL,
        launch_attempts: int = 3,
        post_launch_delay: int = 0,
    ) -> None:
        assert iteration_limit >= 0
        assert launch_attempts > 0
        assert runtime_limit >= 0
        assert time_limit > 0

        LOG.debug("calling adapter.setup()")
        self.adapter.setup(input_path, self.iomanager.server_map)
        LOG.debug("configuring harness (%r)", not no_harness)
        harness = None if no_harness else self.adapter.get_harness()
        LOG.debug("configuring redirects (w/harness: %s)", harness is not None)
        if harness is None:
            self.iomanager.server_map.set_redirect(
                "grz_start", "grz_current_test", required=False
            )
        else:
            assert harness
            self.iomanager.server_map.set_dynamic_response(
                "grz_harness", lambda _: harness, mime_type="text/html", required=False
            )
            self.iomanager.server_map.set_redirect(
                "grz_start", "grz_harness", required=False
            )

        log_limiter = LogOutputLimiter(rate=log_rate)
        # limit relaunch to max iterations if needed
        relaunch = min(self._relaunch, iteration_limit) or self._relaunch
        if self.adapter.remaining is not None:
            assert self.adapter.remaining > 0
            relaunch = min(relaunch, self.adapter.remaining)
        runner = Runner(self.server, self.target, relaunch=relaunch)
        while True:
            self.status.report()
            self.status.iteration += 1
            LOG.debug("- iteration %d -", self.status.iteration)

            if self.target.closed:
                # (re-)launch target
                self.iomanager.purge()
                self.adapter.pre_launch()
                location = runner.location(
                    "/grz_start",
                    self.server.port,
                    close_after=relaunch if harness else None,
                    post_launch_delay=post_launch_delay,
                    scheme=self.server.scheme,
                    time_limit=time_limit if harness else None,
                )
                with self.status.measure("launch"):
                    runner.launch(location, max_retries=launch_attempts, retry_delay=0)
                if post_launch_delay >= 0 and not runner.startup_failure:
                    runner.post_launch(delay=post_launch_delay)
                # TODO: avoid running test case if runner.startup_failure is True
                # especially if it is a hang!

            # create and populate a test case
            current_test = self.generate_testcase()
            # display status
            self.display_status(log_limiter=log_limiter)
            # run test case
            with self.status.measure("execute"):
                result = runner.run(
                    ignore,
                    self.iomanager.server_map,
                    current_test,
                    coverage=self._coverage,
                )
            current_test.duration = result.duration
            # adapter callbacks
            if result.timeout:
                current_test.hang = True
                LOG.debug("calling adapter.on_timeout()")
                self.adapter.on_timeout(current_test, result.served)
            else:
                LOG.debug("calling adapter.on_served()")
                self.adapter.on_served(current_test, result.served)
            if not result.attempted:
                LOG.warning("Test case was not served")
                LOG.debug("ignoring test case since nothing was served")
                if current_test.entry_point not in current_test:
                    LOG.error("Check adapter, test case is missing entry point")
                    raise SessionError("Test case is missing entry point")
                if result.timeout:
                    LOG.warning("Browser hung? Timeout too short? System too busy?")
                elif runner.initial:
                    # since this is the first iteration since the Target launched
                    # something is likely wrong with the Target or Adapter
                    LOG.warning(
                        "Failure detected before running a test case, "
                        "browser build is potentially unstable"
                    )
            else:
                self.iomanager.commit()
            # process results
            if result.status == Result.FOUND:
                LOG.debug("result detected")
                report = self.target.create_report(
                    is_hang=result.timeout,
                    unstable=runner.startup_failure,
                )
                seen, initial = self.status.results.count(
                    report.crash_hash, report.short_signature
                )
                LOG.info(
                    "Result: %s (%s:%s) - %d",
                    report.short_signature,
                    report.major[:8],
                    report.minor[:8],
                    seen,
                )
                if initial or not self.status.results.is_frequent(report.crash_hash):
                    # add target info to test cases
                    for test in self.iomanager.tests:
                        test.assets = dict(self.target.asset_mgr.assets)
                        test.assets_path = self.target.asset_mgr.path
                        test.env_vars = self.target.filtered_environ()
                    self.reporter.submit(self.iomanager.tests, report, force=initial)
                else:
                    # we should always submit the first instance of a result
                    assert seen > 1
                    LOG.info("Result is frequent, skipping submission")
                report.cleanup()
            elif result.status == Result.IGNORED:
                self.status.ignored += 1
                if result.timeout:
                    LOG.info(
                        "Ignored - %d; timeout, idle: %r",
                        self.status.ignored,
                        result.idle,
                    )
                else:
                    LOG.info("Ignored - %d", self.status.ignored)

            if self.adapter.remaining is not None and self.adapter.remaining < 1:
                # all test cases have been replayed
                LOG.info("Replay Complete")
                break

            if iteration_limit and self.status.iteration >= iteration_limit:
                LOG.info("Hit iteration limit (%d)", iteration_limit)
                break

            if result_limit and self.status.results.total >= result_limit:
                LOG.info("Hit result limit (%d)", result_limit)
                break

            if runtime_limit and self.status.runtime >= runtime_limit:
                LOG.info("Hit runtime limit (%ds)", runtime_limit)
                break

            # warn about large browser logs
            self.status.log_size = self.target.log_size()
            if self.status.log_size > self.TARGET_LOG_SIZE_WARN:
                LOG.warning(
                    "Large browser logs: %dMBs", (self.status.log_size / 0x100000)
                )



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from __future__ import annotations

from datetime import datetime, timedelta, timezone
from ipaddress import IPv4Address
from logging import getLogger
from os import getpid
from pathlib import Path
from shutil import rmtree
from tempfile import mkdtemp

from cryptography import x509
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes, serialization
from cryptography.hazmat.primitives.asymmetric.rsa import generate_private_key
from cryptography.x509.oid import NameOID

LOG = getLogger(__name__)


def generate_certificates(cert_dir: Path) -> dict[str, Path]:
    """Generate a root CA and host certificate.

    Credit to https://stackoverflow.com/a/56292132
    """
    root_key = generate_private_key(
        public_exponent=65537,
        key_size=2048,
        backend=default_backend(),
    )
    root_sub = x509.Name([x509.NameAttribute(NameOID.COMMON_NAME, "Grizzly Root CA")])
    root_cert = (
        x509.CertificateBuilder()
        .subject_name(root_sub)
        .issuer_name(root_sub)
        .public_key(root_key.public_key())
        .add_extension(x509.BasicConstraints(ca=True, path_length=None), critical=True)
        .serial_number(x509.random_serial_number())
        .not_valid_before(datetime.now(timezone.utc))
        .not_valid_after(datetime.now(timezone.utc) + timedelta(days=30))
        .sign(root_key, hashes.SHA256(), default_backend())
    )

    # Now we want to generate a cert from that root
    host_key = generate_private_key(
        public_exponent=65537,
        key_size=2048,
        backend=default_backend(),
    )
    host_sub = x509.Name([x509.NameAttribute(NameOID.COMMON_NAME, "Grizzly Test Cert")])
    host_cert = (
        x509.CertificateBuilder()
        .subject_name(host_sub)
        .issuer_name(root_cert.issuer)
        .public_key(host_key.public_key())
        .serial_number(x509.random_serial_number())
        .not_valid_before(datetime.now(timezone.utc))
        .not_valid_after(datetime.now(timezone.utc) + timedelta(days=30))
        .add_extension(
            x509.SubjectAlternativeName(
                [
                    x509.DNSName("localhost"),
                    x509.IPAddress(IPv4Address("127.0.0.1")),
                ]
            ),
            critical=False,
        )
        .sign(root_key, hashes.SHA256(), default_backend())
    )

    root_file = cert_dir / "root.pem"
    root_file.write_bytes(root_cert.public_bytes(serialization.Encoding.PEM))
    cert_file = cert_dir / "host.pem"
    cert_file.write_bytes(host_cert.public_bytes(serialization.Encoding.PEM))
    key_file = cert_dir / "host.key"
    key_file.write_bytes(
        host_key.private_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PrivateFormat.TraditionalOpenSSL,
            encryption_algorithm=serialization.NoEncryption(),
        )
    )

    return {"root": root_file, "host": cert_file, "key": key_file}


class CertificateBundle:
    """Contains root CA, host CA and private key files."""

    def __init__(self, path: Path, root: Path, host: Path, key: Path) -> None:
        self._base = path
        self.root = root
        self.host = host
        self.key = key

    @classmethod
    def create(cls, path: Path | None = None) -> CertificateBundle:
        """Create certificate files.

        Args:
            path: Location to store generated files.

        Returns:
            CertificateBundle
        """
        if path is None:
            path = Path(mkdtemp(prefix=f"sapphire_certs_{getpid()}_"))
        certs = generate_certificates(path)
        return cls(path, certs["root"], certs["host"], certs["key"])

    def cleanup(self) -> None:
        """Remove certificate files.

        Args:
            None

        Returns:
            None
        """
        LOG.debug("removing certificate path (%s)", self._base)
        rmtree(self._base, ignore_errors=True)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""Unit tests for `grizzly.replay.bugzilla`"""
from base64 import b64encode

from bugsy import Bug

from .args import ReplayFuzzBugzillaArgs
from .bugzilla import main


def test_crash_main_no_bug(mocker, tmp_path):
    """test main() - failed to get bug"""
    bugzilla = mocker.patch("grizzly.replay.bugzilla.BugzillaBug", autospec=True)
    bugzilla.load.return_value = None
    file = tmp_path / "file"
    file.touch()
    args = ReplayFuzzBugzillaArgs().parse_args([str(file), "123"])
    assert main(args) == 1


def test_crash_main(mocker, tmp_path):
    """test main()"""
    fake_bugsy = mocker.patch("grizzly.common.bugzilla.Bugsy", autospec=True)
    fake_replay = mocker.patch("grizzly.replay.bugzilla.ReplayManager", autospec=True)
    fake_replay.main.return_value = 0
    file = tmp_path / "file"
    file.touch()

    # no test cases
    fake_bugsy.return_value.get.return_value = mocker.MagicMock(spec=Bug, id=123)
    args = ReplayFuzzBugzillaArgs().parse_args([str(file), "123"])
    assert main(args) == 1
    assert fake_bugsy.return_value.get.call_count == 1
    fake_bugsy.reset_mock()

    # test case and asset
    bugsy_bug = mocker.Mock(spec=Bug, id=123)
    bugsy_bug.get_attachments.return_value = [
        mocker.Mock(is_obsolete=False, data=b64encode(b"foo"), file_name="test.html"),
        mocker.Mock(is_obsolete=False, data=b64encode(b"foo"), file_name="prefs.js"),
    ]
    fake_bugsy.return_value.get.return_value = bugsy_bug
    args = ReplayFuzzBugzillaArgs().parse_args([str(file), "123"])
    assert main(args) == 0
    assert fake_bugsy.return_value.get.call_count == 1
    assert args.input != "123"
    assert args.asset
    assert "prefs" in args.asset[-1]



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
"""test Grizzly Reporter"""
# pylint: disable=protected-access
from pathlib import Path

from pytest import mark, raises

from .report import Report
from .reporter import (
    FailedLaunchReporter,
    FilesystemReporter,
    FuzzManagerReporter,
    Quality,
    Reporter,
)
from .storage import TestCase


def _create_crash_log(log_path):
    with log_path.open("w") as log_fp:
        log_fp.write("==1==ERROR: AddressSanitizer: SEGV on unknown address 0x0")
        log_fp.write(" (pc 0x0 bp 0x0 sp 0x0 T0)\n")
        log_fp.write("    #0 0xbad000 in foo /file1.c:123:234\n")
        log_fp.write("    #1 0x1337dd in bar /file2.c:1806:19")


@mark.parametrize(
    "display_logs, is_hang",
    [
        # do not display report
        (False, False),
        # display report
        (True, False),
        # display report (hang)
        (True, True),
    ],
)
def test_reporter_01(mocker, tmp_path, display_logs, is_hang):
    """test creating a simple Reporter"""

    class SimpleReporter(Reporter):
        def _pre_submit(self, report):
            pass

        def _post_submit(self):
            pass

        def _submit_report(self, report, test_cases, force):
            pass

    (tmp_path / "log_stderr.txt").write_bytes(b"log msg")
    report = mocker.Mock(
        spec_set=Report, is_hang=is_hang, preferred=tmp_path / "log_stderr.txt"
    )
    reporter = SimpleReporter()
    reporter.display_logs = display_logs
    reporter.submit([], report)
    assert report.cleanup.call_count == 1


def test_filesystem_reporter_01(tmp_path):
    """test FilesystemReporter without testcases"""
    log_path = tmp_path / "logs"
    log_path.mkdir()
    (log_path / "log_stderr.txt").write_bytes(b"STDERR log")
    (log_path / "log_stdout.txt").write_bytes(b"STDOUT log")
    _create_crash_log(tmp_path / "log_asan_blah.txt")
    report_path = tmp_path / "reports"
    report_path.mkdir()
    reporter = FilesystemReporter(report_path)
    assert reporter.submit([], Report(log_path, Path("bin")))
    buckets = tuple(report_path.iterdir())
    # check major bucket
    assert len(buckets) == 1
    assert buckets[0].is_dir()
    # check log path exists
    log_dirs = tuple(buckets[0].iterdir())
    assert len(log_dirs) == 1
    assert log_dirs[0].is_dir()
    assert "_logs" in str(log_dirs[0])


def test_filesystem_reporter_02(tmp_path, mocker):
    """test FilesystemReporter with testcases"""
    log_path = tmp_path / "logs"
    log_path.mkdir()
    (log_path / "log_stderr.txt").write_bytes(b"STDERR log")
    (log_path / "log_stdout.txt").write_bytes(b"STDOUT log")
    _create_crash_log(log_path / "log_asan_blah.txt")
    tests = [mocker.Mock(spec_set=TestCase, timestamp=x) for x in range(10)]
    report_path = tmp_path / "reports"
    assert not report_path.exists()
    reporter = FilesystemReporter(report_path)
    reporter.submit(tests, Report(log_path, Path("bin")))
    assert not log_path.exists()
    assert report_path.exists()
    assert len(tuple(report_path.iterdir())) == 1
    assert all(x.dump.call_count == 1 for x in tests)
    # call report a 2nd time
    log_path.mkdir()
    (log_path / "log_stderr.txt").write_bytes(b"STDERR log")
    (log_path / "log_stdout.txt").write_bytes(b"STDOUT log")
    tests = [mocker.Mock(spec_set=TestCase, timestamp=1) for _ in range(2)]
    reporter.submit(tests, Report(log_path, Path("bin")))
    assert all(x.dump.call_count == 1 for x in tests)
    assert len(tuple(report_path.iterdir())) == 2
    assert len(tuple(report_path.glob("NO_STACK"))) == 1


def test_filesystem_reporter_03(tmp_path):
    """test FilesystemReporter disk space failsafe"""
    log_path = tmp_path / "logs"
    log_path.mkdir()
    (log_path / "log_stderr.txt").write_bytes(b"STDERR log")
    (log_path / "log_stdout.txt").write_bytes(b"STDOUT log")
    reporter = FilesystemReporter(tmp_path / "reports")
    reporter.min_space = 2**50
    with raises(RuntimeError, match="Low disk space: "):
        reporter.submit([], Report(log_path, Path("bin")))


def test_filesystem_reporter_04(mocker, tmp_path):
    """test FilesystemReporter w/o major bucket"""
    fake_report = tmp_path / "fake_report"
    fake_report.mkdir()
    report = mocker.Mock(spec_set=Report, path=fake_report, prefix="test_prefix")
    reporter = FilesystemReporter(tmp_path / "dst", major_bucket=False)
    reporter.submit([], report)
    assert not fake_report.is_dir()
    assert not report.major.call_count
    assert any((tmp_path / "dst").glob("test_prefix_logs"))


@mark.parametrize(
    "tests, frequent, force, sig_cache",
    [
        # report - without test
        (False, False, False, True),
        # report - with test
        (True, False, False, True),
        # report - frequent
        (True, True, False, True),
        # report - forced frequent
        (True, True, True, True),
        # report - missing sigCacheDir
        (False, False, False, False),
    ],
)
def test_fuzzmanager_reporter_01(mocker, tmp_path, tests, frequent, force, sig_cache):
    """test FuzzManagerReporter.submit()"""
    mocker.patch("grizzly.common.reporter.Path.cwd", return_value=tmp_path)
    mocker.patch("grizzly.common.reporter.getenv", autospec=True, return_value="0")
    fake_collector = mocker.patch("grizzly.common.reporter.Collector", autospec=True)
    fake_collector.return_value.search.return_value = (
        None,
        {"frequent": frequent, "shortDescription": "[@ test]"},
    )
    fake_collector.return_value.sigCacheDir = str(tmp_path) if sig_cache else None
    fake_collector.return_value.tool = "fake-tool"
    log_path = tmp_path / "log_path"
    log_path.mkdir()
    (log_path / "log_ffp_worker_blah.txt").touch()
    (log_path / "log_stderr.txt").touch()
    (log_path / "log_stdout.txt").touch()
    (log_path / "rr-traces").mkdir()
    (tmp_path / "screenlog.0").touch()
    fake_test = mocker.Mock(
        spec_set=TestCase,
        adapter_name="adapter",
        env_vars={"TEST": "1"},
        input_fname="input",
        timestamp=1234.5678,
    )
    test_cases = []
    if tests:
        test_cases.append(fake_test)
    reporter = FuzzManagerReporter("fake-tool")
    reporter.submit(
        test_cases,
        Report(log_path, Path("bin"), is_hang=True, unstable=True),
        force=force,
    )
    assert not log_path.is_dir()
    assert fake_collector.call_args == ({"tool": "fake-tool"},)
    if frequent and not force:
        assert fake_collector.return_value.submit.call_count == 0
        assert fake_test.dump.call_count == 0
    else:
        assert fake_collector.return_value.submit.call_count == 1
        if tests:
            assert fake_test.dump.call_count == 1
    if fake_collector.return_value.submit.call_count:
        if tests:
            assert (
                fake_collector.return_value.submit.call_args.kwargs["testCaseQuality"]
                != Quality.NO_TESTCASE
            )
        else:
            assert (
                fake_collector.return_value.submit.call_args.kwargs["testCaseQuality"]
                == Quality.NO_TESTCASE
            )


@mark.parametrize(
    "tool, sanitized",
    [
        # no whitespace
        ("test-tool", "test-tool"),
        # whitespace
        ("test tool", "test-tool"),
        # whitespace
        ("test\t\r\n\vtool", "test-tool"),
        # mixed case
        ("TeSt-ToOl", "test-tool"),
    ],
)
def test_fuzzmanager_reporter_03(tool, sanitized):
    """test FuzzManagerReporter() sanitizing tool"""
    assert FuzzManagerReporter(tool).tool == sanitized


def test_failed_launch_reporter_01(mocker, tmp_path):
    """test FailedLaunchReporter()"""
    gtp = tmp_path / "grizzly"
    gtp.mkdir()
    mocker.patch("grizzly.common.reporter.grz_tmp", autospec=True, return_value=gtp)
    report_path = tmp_path / "report"
    report_path.mkdir()
    (report_path / "log_stderr.txt").write_bytes(b"STDERR log")
    (report_path / "log_stdout.txt").write_bytes(b"STDOUT log")
    FailedLaunchReporter().submit([], Report(report_path, Path("bin")))
    assert any(gtp.glob("*_logs"))



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# pylint: disable=protected-access
"""Unit tests for `grizzly.reduce.reduce`.
"""
from collections import namedtuple
from functools import partial, wraps
from itertools import count
from logging import getLogger
from pathlib import Path

from pytest import mark, param, raises

from sapphire import Sapphire

from ..common.reporter import Quality, Report
from ..common.storage import TestCase
from ..replay import ReplayResult
from ..target import AssetManager, Target
from . import ReduceManager
from .exceptions import NotReproducible
from .strategies import Strategy

LOG = getLogger(__name__)
pytestmark = mark.usefixtures(
    "reporter_sequential_strftime",
    "tmp_path_status_db_reduce",
)


def _fake_save_logs_foo(result_logs):
    """write fake log data to disk"""
    (result_logs / "log_stderr.txt").write_text("STDERR log\n")
    (result_logs / "log_stdout.txt").write_text("STDOUT log\n")
    (result_logs / "log_asan_blah.txt").write_text(
        "==1==ERROR: AddressSanitizer: "
        "SEGV on unknown address 0x0 (pc 0x0 bp 0x0 sp 0x0 T0)\n"
        "    #0 0xbad000 in foo /file1.c:123:234\n"
        "    #1 0x1337dd in bar /file2.c:1806:19\n"
    )


def _fake_save_logs_bar(result_logs):
    """write fake log data to disk"""
    (result_logs / "log_stderr.txt").write_text("STDERR log\n")
    (result_logs / "log_stdout.txt").write_text("STDOUT log\n")
    (result_logs / "log_asan_blah.txt").write_text(
        "==1==ERROR: AddressSanitizer: "
        "SEGV on unknown address 0x0 (pc 0x0 bp 0x0 sp 0x0 T0)\n"
        "    #0 0xbad000 in bar /file1.c:123:234\n"
        "    #1 0x1337dd in foo /file2.c:1806:19\n"
    )


@mark.parametrize(
    "harness_last_crashes,harness_crashes,no_harness_crashes,expected_repeat,"
    "expected_min_crashes,use_harness,result_harness",
    [
        param(
            11,
            None,
            None,
            ReduceManager.ANALYSIS_PERFECT_MIN_CRASHES,
            ReduceManager.ANALYSIS_PERFECT_MIN_CRASHES,
            True,
            True,
            id="perfect with harness",
        ),
        param(
            None,
            None,
            11,
            ReduceManager.ANALYSIS_PERFECT_MIN_CRASHES,
            ReduceManager.ANALYSIS_PERFECT_MIN_CRASHES,
            False,
            False,
            id="perfect, use_harness=False",
        ),
        param(
            1,
            11,
            None,
            ReduceManager.ANALYSIS_PERFECT_MIN_CRASHES,
            ReduceManager.ANALYSIS_PERFECT_MIN_CRASHES,
            True,
            True,
            id="1/11 with harness last test, perfect all tests",
        ),
        param(0, 10, None, 2, 1, True, True, id="10/11 with harness"),
        param(0, 9, None, 2, 1, True, True, id="9/11 with harness"),
        param(0, 8, None, 3, 1, True, True, id="8/11 with harness"),
        param(0, 7, None, 3, 1, True, True, id="7/11 with harness"),
        param(0, 6, None, 4, 1, True, True, id="6/11 with harness"),
        param(0, 5, 0, 5, 1, True, True, id="5/11 with harness"),
        param(0, 5, 1, 5, 1, True, True, id="5/11 with harness, 1/11 without"),
        param(0, 5, 2, 5, 1, True, True, id="5/11 with harness, 2/11 without"),
        param(0, 5, 3, 5, 1, True, True, id="5/11 with harness, 3/11 without"),
        param(0, 5, 4, 5, 1, True, True, id="5/11 with harness, 4/11 without"),
        param(0, 5, 5, 5, 1, True, True, id="5/11 with harness, 5/11 without"),
        param(0, 5, 6, 5, 1, True, True, id="5/11 with harness, 6/11 without"),
        param(0, 5, 7, 5, 1, True, True, id="5/11 with harness, 7/11 without"),
        param(0, 5, 8, 3, 1, True, False, id="5/11 with harness, 8/11 without"),
        param(0, 5, 9, 2, 1, True, False, id="5/11 with harness, 9/11 without"),
        param(0, 5, 10, 2, 1, True, False, id="5/11 with harness, 10/11 without"),
        param(
            0,
            5,
            11,
            ReduceManager.ANALYSIS_PERFECT_MIN_CRASHES,
            ReduceManager.ANALYSIS_PERFECT_MIN_CRASHES,
            True,
            False,
            id="5/11 with harness, perfect without",
        ),
        param(0, 4, 0, 7, 1, True, True, id="4/11 with harness"),
        param(0, 4, 1, 7, 1, True, True, id="4/11 with harness, 1/11 without"),
        param(0, 4, 2, 7, 1, True, True, id="4/11 with harness, 2/11 without"),
        param(0, 4, 3, 7, 1, True, True, id="4/11 with harness, 3/11 without"),
        param(0, 4, 4, 7, 1, True, True, id="4/11 with harness, 4/11 without"),
        param(0, 4, 5, 7, 1, True, True, id="4/11 with harness, 5/11 without"),
        param(0, 4, 6, 4, 1, True, False, id="4/11 with harness, 6/11 without"),
        param(0, 4, 7, 3, 1, True, False, id="4/11 with harness, 7/11 without"),
        param(0, 4, 8, 3, 1, True, False, id="4/11 with harness, 8/11 without"),
        param(0, 4, 9, 2, 1, True, False, id="4/11 with harness, 9/11 without"),
        param(0, 4, 10, 2, 1, True, False, id="4/11 with harness, 10/11 without"),
        param(
            0,
            4,
            11,
            ReduceManager.ANALYSIS_PERFECT_MIN_CRASHES,
            ReduceManager.ANALYSIS_PERFECT_MIN_CRASHES,
            True,
            False,
            id="4/11 with harness, perfect without",
        ),
        param(0, 3, 0, 10, 1, True, True, id="3/11 with harness"),
        param(0, 3, 1, 10, 1, True, True, id="3/11 with harness, 1/11 without"),
        param(0, 3, 2, 10, 1, True, True, id="3/11 with harness, 2/11 without"),
        param(0, 3, 3, 10, 1, True, True, id="3/11 with harness, 3/11 without"),
        param(0, 3, 4, 10, 1, True, True, id="3/11 with harness, 4/11 without"),
        param(0, 3, 5, 5, 1, True, False, id="3/11 with harness, 5/11 without"),
        param(0, 3, 6, 4, 1, True, False, id="3/11 with harness, 6/11 without"),
        param(0, 3, 7, 3, 1, True, False, id="3/11 with harness, 7/11 without"),
        param(0, 3, 8, 3, 1, True, False, id="3/11 with harness, 8/11 without"),
        param(0, 3, 9, 2, 1, True, False, id="3/11 with harness, 9/11 without"),
        param(0, 3, 10, 2, 1, True, False, id="3/11 with harness, 10/11 without"),
        param(
            0,
            3,
            11,
            ReduceManager.ANALYSIS_PERFECT_MIN_CRASHES,
            ReduceManager.ANALYSIS_PERFECT_MIN_CRASHES,
            True,
            False,
            id="3/11 with harness, perfect without",
        ),
        param(0, 2, 0, 15, 1, True, True, id="2/11 with harness"),
        param(0, 2, 1, 15, 1, True, True, id="2/11 with harness, 1/11 without"),
        param(0, 2, 2, 15, 1, True, True, id="2/11 with harness, 2/11 without"),
        param(0, 2, 3, 10, 1, True, False, id="2/11 with harness, 3/11 without"),
        param(0, 2, 4, 7, 1, True, False, id="2/11 with harness, 4/11 without"),
        param(0, 2, 5, 5, 1, True, False, id="2/11 with harness, 5/11 without"),
        param(0, 2, 6, 4, 1, True, False, id="2/11 with harness, 6/11 without"),
        param(0, 2, 7, 3, 1, True, False, id="2/11 with harness, 7/11 without"),
        param(0, 2, 8, 3, 1, True, False, id="2/11 with harness, 8/11 without"),
        param(0, 2, 9, 2, 1, True, False, id="2/11 with harness, 9/11 without"),
        param(0, 2, 10, 2, 1, True, False, id="2/11 with harness, 10/11 without"),
        param(
            0,
            2,
            11,
            ReduceManager.ANALYSIS_PERFECT_MIN_CRASHES,
            ReduceManager.ANALYSIS_PERFECT_MIN_CRASHES,
            True,
            False,
            id="2/11 with harness, perfect without",
        ),
        param(0, 1, 0, 32, 1, True, True, id="1/11 with harness"),
        param(0, 1, 1, 32, 1, True, True, id="1/11 with harness, 1/11 without"),
        param(0, 1, 2, 15, 1, True, False, id="1/11 with harness, 2/11 without"),
        param(0, 1, 3, 10, 1, True, False, id="1/11 with harness, 3/11 without"),
        param(0, 1, 4, 7, 1, True, False, id="1/11 with harness, 4/11 without"),
        param(0, 1, 5, 5, 1, True, False, id="1/11 with harness, 5/11 without"),
        param(0, 1, 6, 4, 1, True, False, id="1/11 with harness, 6/11 without"),
        param(0, 1, 7, 3, 1, True, False, id="1/11 with harness, 7/11 without"),
        param(0, 1, 8, 3, 1, True, False, id="1/11 with harness, 8/11 without"),
        param(0, 1, 9, 2, 1, True, False, id="1/11 with harness, 9/11 without"),
        param(0, 1, 10, 2, 1, True, False, id="1/11 with harness, 10/11 without"),
        param(
            0,
            1,
            11,
            ReduceManager.ANALYSIS_PERFECT_MIN_CRASHES,
            ReduceManager.ANALYSIS_PERFECT_MIN_CRASHES,
            True,
            False,
            id="1/11 with harness, perfect without",
        ),
        param(
            0,
            0,
            11,
            ReduceManager.ANALYSIS_PERFECT_MIN_CRASHES,
            ReduceManager.ANALYSIS_PERFECT_MIN_CRASHES,
            True,
            False,
            id="perfect without harness",
        ),
        param(0, 0, 10, 2, 1, True, False, id="10/11 without harness"),
        param(0, 0, 9, 2, 1, True, False, id="9/11 without harness"),
        param(0, 0, 8, 3, 1, True, False, id="8/11 without harness"),
        param(0, 0, 7, 3, 1, True, False, id="7/11 without harness"),
        param(0, 0, 6, 4, 1, True, False, id="6/11 without harness"),
        param(0, 0, 5, 5, 1, True, False, id="5/11 without harness"),
        param(0, 0, 4, 7, 1, True, False, id="4/11 without harness"),
        param(0, 0, 3, 10, 1, True, False, id="3/11 without harness"),
        param(0, 0, 2, 15, 1, True, False, id="2/11 without harness"),
        param(0, 0, 1, 32, 1, True, False, id="1/11 without harness"),
        param(
            1,
            0,
            0,
            32,
            1,
            True,
            True,
            id="last test repros (1/11), all tests are tried, without_harness is tried",
        ),
        param(
            11,
            None,
            None,
            ReduceManager.ANALYSIS_PERFECT_MIN_CRASHES,
            ReduceManager.ANALYSIS_PERFECT_MIN_CRASHES,
            True,
            True,
            id="last test repros (11/11), no other rounds are tried",
        ),
        param(
            0,
            1,
            0,
            32,
            1,
            True,
            True,
            id="last test doesn't repro, harness+all tests is (1/11), "
            "without harness is tried",
        ),
        param(
            0,
            11,
            None,
            ReduceManager.ANALYSIS_PERFECT_MIN_CRASHES,
            ReduceManager.ANALYSIS_PERFECT_MIN_CRASHES,
            True,
            True,
            id="last test doesn't repro, harness+all tests is (11/11), "
            "skip without harness",
        ),
    ],
)
def test_analysis(
    mocker,
    tmp_path,
    harness_last_crashes,
    harness_crashes,
    no_harness_crashes,
    expected_repeat,
    expected_min_crashes,
    use_harness,
    result_harness,
):
    """test that analysis sets reasonable params"""
    replayer = mocker.patch("grizzly.reduce.core.ReplayManager", autospec=True)
    replayer = replayer.return_value.__enter__.return_value
    crashes = []
    expected_iters = 0
    if harness_last_crashes is not None:
        crashes += [False] * (11 - harness_last_crashes) + [True] * harness_last_crashes
        expected_iters += 1
    if harness_crashes is not None:
        crashes += [True] * harness_crashes + [False] * (11 - harness_crashes)
        expected_iters += 1
    if no_harness_crashes is not None:
        crashes += [False] * (11 - no_harness_crashes) + [True] * no_harness_crashes
        expected_iters += 1

    def replay_run(_tests, _time_limit, **kw):
        results = []
        repeat = kw["repeat"]
        assert repeat <= len(crashes)
        assert not kw["exit_early"]
        for _ in range(repeat):
            kw["on_iteration_cb"]()
            LOG.debug("interesting: %r", crashes[0])
            if crashes.pop(0):
                log_path = tmp_path / f"crash{replayer.run.call_count}_logs"
                log_path.mkdir(exist_ok=True)
                _fake_save_logs_foo(log_path)
                report = Report(log_path, Path("bin"))
                results.append(ReplayResult(report, [], True))
        return results

    replayer.run.side_effect = replay_run

    with TestCase("test.html", "test-adapter") as test:
        test.add_from_bytes(b"1", test.entry_point)
        test.dump(tmp_path / "src1", include_details=True)
    tests = [test.load(tmp_path / "src1")]
    if harness_last_crashes is not None:
        with TestCase("test.html", "test-adapter") as test:
            test.add_from_bytes(b"2", test.entry_point)
            test.dump(tmp_path / "src2", include_details=True)
        tests.append(test.load(tmp_path / "src2"))
    log_path = tmp_path / "logs"

    with ReduceManager(
        set(),
        mocker.Mock(spec_set=Sapphire, timeout=30),
        mocker.Mock(spec_set=Target),
        tests,
        [],
        log_path,
        use_harness=use_harness,
    ) as mgr:
        repeat, min_crashes = mgr.run_reliability_analysis()

        observed = {
            "replay_iters": replayer.run.call_count,
            "repeat": repeat,
            "min_crashes": min_crashes,
            "use_harness": mgr._use_harness,
            "launch_iters": mgr._status.iterations,
        }
        expected = {
            "replay_iters": expected_iters,
            "repeat": expected_repeat,
            "min_crashes": expected_min_crashes,
            "use_harness": result_harness,
            "launch_iters": expected_iters * 11,
        }
    assert observed == expected


def _ignore_arg(func):
    """Function wrapper that simply ignores 1 argument"""

    @wraps(func)
    def wrapped(_):
        return func()

    return wrapped


ReproTestParams = namedtuple(
    "ReproTestParams",
    "original, strategies, detect_failure, interesting_str, is_expected,"
    "expected_run_calls, n_reports, reports, n_other, other_reports, result",
)


@mark.parametrize(
    ReproTestParams._fields,
    [
        # no repro
        ReproTestParams(
            original=b"123456\n",
            strategies=["check"],
            detect_failure=lambda _: False,
            interesting_str="%r",
            is_expected=None,
            expected_run_calls=1,
            n_reports=0,
            reports=None,
            n_other=0,
            other_reports=None,
            result=NotReproducible,
        ),
        # reproduces, one strategy, no reduction works
        ReproTestParams(
            original=b"1\n2\n3\n",
            strategies=["check", "lines"],
            detect_failure=lambda contents: contents == "1\n2\n3\n",
            interesting_str="%r == '1\n2\n3\n'",
            is_expected=lambda _: True,
            expected_run_calls=6,
            n_reports=1,
            reports={"1\n2\n3\n"},
            n_other=0,
            other_reports=None,
            result=0,
        ),
        # reproduces, one strategy, some reduction works
        ReproTestParams(
            original=b"odd\neven\n" * 3,
            strategies=["check", "lines"],
            detect_failure=lambda contents: sum(
                1 for line in contents.splitlines() if line == "odd"
            )
            == 3,
            interesting_str="%r contains 3x 'odd'",
            is_expected=lambda _: True,
            expected_run_calls=8,
            n_reports=2,
            reports={"odd\neven\n" * 3, "odd\n" * 3},
            n_other=0,
            other_reports=None,
            result=0,
        ),
        # reproduces, one strategy, reduction only finds other
        ReproTestParams(
            original=b"1\n2\n3\n",
            strategies=["check", "lines"],
            detect_failure=bool,
            interesting_str="%r != ''",
            is_expected=lambda contents: contents == "1\n2\n3\n",
            expected_run_calls=6,
            n_reports=1,
            reports={"1\n2\n3\n"},
            n_other=1,
            other_reports={"1\n"},
            result=0,
        ),
        # reproduces, 50% iterations work, 1 iteration produces other sig
        ReproTestParams(
            original=b"1\n2\n3\n4\n5\n6\n",
            strategies=["check", "lines"],
            detect_failure=lambda contents: set(contents.splitlines()) >= set("135"),
            interesting_str="%r contains {'1', '3', '5'}",
            is_expected=lambda contents: contents != "1\n3\n5\n",
            expected_run_calls=13,
            n_reports=2,
            reports={"1\n2\n3\n4\n5\n6\n", "1\n2\n3\n5\n"},
            n_other=1,
            other_reports={"1\n3\n5\n"},
            result=0,
        ),
        # reproduces, two strategies, 1st no reduction, 2nd 50% reduction
        ReproTestParams(
            original=b"A1\nA2\nA3\nA4\nA5\nA6\n",
            strategies=["check", "lines", "chars"],
            detect_failure=(
                lambda contents: (
                    len(contents.splitlines()) == 6
                    and set(contents.splitlines()) >= {"A1", "A3", "A5"}
                )
            ),
            interesting_str="%r contains {'A1', 'A3', 'A5'} and len() == 6",
            is_expected=lambda _: True,
            expected_run_calls=43,
            # lines found nothing, only check and chars should report
            n_reports=2,
            reports={"A1\nA2\nA3\nA4\nA5\nA6\n", "A1\n\nA3\n\nA5\nA"},
            n_other=0,
            other_reports=None,
            result=0,
        ),
        # reproduces, two strategies, 1st 50% reduction, 2nd no reduction
        ReproTestParams(
            original=b"A1\nA2\nA3\nA4\nA5\nA6\n",
            strategies=["check", "lines", "chars"],
            detect_failure=(
                lambda contents: (
                    set(contents.splitlines(keepends=True)) >= {"A1\n", "A3\n", "A5\n"}
                )
            ),
            interesting_str="%r contains {'A1\\n', 'A3\\n', 'A5\\n'}",
            is_expected=lambda _: True,
            expected_run_calls=34,
            # lines found nothing, only check and chars should report
            n_reports=2,
            reports={"A1\nA2\nA3\nA4\nA5\nA6\n", "A1\nA3\nA5\n"},
            n_other=0,
            other_reports=None,
            result=0,
        ),
        # reproduces, two strategies, reduce only produces other sig
        ReproTestParams(
            original=b"1\n2\n3\n",
            strategies=["check", "lines", "chars"],
            detect_failure=bool,
            interesting_str="%r != ''",
            is_expected=lambda contents: contents == "1\n2\n3\n",
            expected_run_calls=12,
            n_reports=1,
            reports={"1\n2\n3\n"},
            n_other=2,
            other_reports={
                "1\n",
                "1\n2\n3",
            },
            result=0,
        ),
        # reproduces, one strategy, testcase reduces to 0
        ReproTestParams(
            original=b"1\n2\n3\n",
            strategies=["check", "lines"],
            detect_failure=lambda _: True,
            interesting_str="%r is anything, incl. empty",
            is_expected=lambda _: True,
            expected_run_calls=3,
            n_reports=2,
            reports={"1\n2\n3\n", ""},
            n_other=0,
            other_reports=None,
            result=0,
        ),
        # reproduces, two strategies, 1st no reduce, 2nd testcase reduces to 0
        ReproTestParams(
            original=b"1\n2\n3\n",
            strategies=["check", "lines", "lines"],
            detect_failure=_ignore_arg(
                partial([True, False, False, False, False, True, True, True].pop, 0)
            ),
            interesting_str="%r is anything, only in second strategy",
            is_expected=lambda _: True,
            expected_run_calls=8,
            n_reports=2,
            reports={"1\n2\n3\n", ""},
            n_other=0,
            other_reports=None,
            result=0,
        ),
    ],
)
@mark.usefixtures("reporter_sequential_strftime")
def test_repro(
    mocker,
    tmp_path,
    original,
    strategies,
    detect_failure,
    interesting_str,
    is_expected,
    expected_run_calls,
    n_reports,
    reports,
    n_other,
    other_reports,
    result,
):
    """test ReduceManager, difference scenarios produce correct expected/other
    results"""
    mocker.patch("grizzly.reduce.strategies.lithium._contains_dd", return_value=True)
    replayer = mocker.patch("grizzly.reduce.core.ReplayManager", autospec=True)
    replayer = replayer.return_value

    def replay_run(testcases, _time_limit, **kw):
        kw["on_iteration_cb"]()
        for test in testcases:
            contents = test["test.html"].read_text()
            # pylint: disable=logging-not-lazy
            LOG.debug("interesting if " + interesting_str, contents)
            if detect_failure(contents):
                log_path = tmp_path / f"crash{replayer.run.call_count}_logs"
                log_path.mkdir()
                expected = is_expected(contents)
                if expected:
                    _fake_save_logs_foo(log_path)
                else:
                    _fake_save_logs_bar(log_path)
                report = Report(log_path, Path("bin"))
                return [ReplayResult(report, [], expected)]
        return []

    replayer.run.side_effect = replay_run

    with TestCase("test.html", "test-adapter") as test:
        test.add_from_bytes(original, test.entry_point)
        test.dump(tmp_path / "src")
    tests = [TestCase.load(tmp_path / "src")]

    log_path = tmp_path / "logs"

    target = mocker.Mock(spec_set=Target)
    target.filtered_environ.return_value = {}
    target.asset_mgr = mocker.Mock(spec_set=AssetManager)
    with ReduceManager(
        set(),
        mocker.Mock(spec_set=Sapphire, timeout=30),
        target,
        tests,
        strategies,
        log_path,
        use_analysis=False,
    ) as mgr:
        if isinstance(result, type) and issubclass(result, BaseException):
            with raises(result):
                mgr.run()
        else:
            assert mgr.run() == result

    expected_dirs = set()
    if n_reports:
        expected_dirs.add(log_path / "reports")
    if n_other:
        expected_dirs.add(log_path / "other_reports")
    if expected_dirs:
        assert set(log_path.iterdir()) == expected_dirs
    if n_reports:
        tests = {test.read_text() for test in log_path.glob("reports/*-0/test.html")}
        assert tests == reports
        assert sum(1 for _ in (log_path / "reports").iterdir()) == n_reports * 2, list(
            (log_path / "reports").iterdir()
        )
    if n_other:
        other_tests = {
            test.read_text() for test in log_path.glob("other_reports/*-0/test.html")
        }
        assert other_tests == other_reports
        assert (
            sum(1 for _ in (log_path / "other_reports").iterdir()) == n_other * 2
        ), list((log_path / "other_reports").iterdir())
    assert replayer.run.call_count == expected_run_calls


def test_report_01(mocker, tmp_path):
    """test that report is called with --report-period is set"""
    mocker.patch("grizzly.reduce.core.time", side_effect=count())

    replayer = mocker.patch("grizzly.reduce.core.ReplayManager", autospec=True)
    replayer = replayer.return_value

    def replay_run(_tests, _time_limit, **kw):
        kw["on_iteration_cb"]()
        if replayer.run.call_count in {20, 40}:
            log_path = tmp_path / f"crash{replayer.run.call_count}_logs"
            log_path.mkdir()
            _fake_save_logs_foo(log_path)
            report = Report(log_path, Path("bin"))
            return [ReplayResult(report, [], True)]
        return []

    replayer.run.side_effect = replay_run

    (tmp_path / "test.html").touch()
    testcase = TestCase.load(tmp_path / "test.html")
    assert testcase
    log_path = tmp_path / "logs"

    fake_strat = mocker.MagicMock(spec_set=Strategy)
    fake_strat.return_value.name = "fake"

    def fake_iter():
        for count_ in range(1, 61):
            LOG.debug("fake_iter() %d", count_)
            (tmp_path / "test.html").write_text(str(count_))
            testcases = [TestCase.load(tmp_path / "test.html")]
            assert testcases
            yield testcases

    fake_strat.return_value.__iter__.side_effect = fake_iter
    mocker.patch("grizzly.reduce.core.STRATEGIES", new={"fake": fake_strat})

    target = mocker.Mock(spec_set=Target)
    target.filtered_environ.return_value = {}
    target.asset_mgr = mocker.Mock(spec_set=AssetManager)
    with ReduceManager(
        set(),
        mocker.Mock(spec_set=Sapphire, timeout=30),
        target,
        [testcase],
        ["fake"],
        log_path,
        use_analysis=False,
        report_period=30,
    ) as mgr:
        assert mgr.run() == 0

    # should be 2 reports: one made at time=30 (for crash on 20th iter),
    # and one at time=60 (for crash on 40th iter)
    n_reports = 2
    reports = {"20", "40"}
    assert replayer.run.call_count == 60
    expected_dirs = {log_path / "reports"}
    assert set(log_path.iterdir()) == expected_dirs
    tests = {test.read_text() for test in log_path.glob("reports/*-0/test.html")}
    assert tests == reports
    assert sum(1 for _ in (log_path / "reports").iterdir()) == n_reports * 2, list(
        (log_path / "reports").iterdir()
    )


def test_report_02(mocker, tmp_path):
    """test that report is called when KeyboardInterrupt occurs"""
    mocker.patch("grizzly.reduce.core.time", side_effect=count())

    replayer = mocker.patch("grizzly.reduce.core.ReplayManager", autospec=True)
    replayer = replayer.return_value

    def replay_run(_tests, _time_limit, **kw):
        kw["on_iteration_cb"]()
        if replayer.run.call_count in {10, 20}:
            log_path = tmp_path / f"crash{replayer.run.call_count}_logs"
            log_path.mkdir()
            _fake_save_logs_foo(log_path)
            report = Report(log_path, Path("bin"))
            return [ReplayResult(report, [], True)]
        return []

    replayer.run.side_effect = replay_run

    (tmp_path / "test.html").touch()
    testcase = TestCase.load(tmp_path / "test.html")
    assert testcase
    log_path = tmp_path / "logs"

    fake_strat = mocker.MagicMock(spec_set=Strategy)
    fake_strat.return_value.name = "fake"

    def fake_iter():
        for count_ in range(1, 31):
            LOG.debug("fake_iter() %d", count_)
            (tmp_path / "test.html").write_text(str(count_))
            testcases = [TestCase.load(tmp_path / "test.html")]
            assert testcases
            yield testcases
        raise KeyboardInterrupt()

    fake_strat.return_value.__iter__.side_effect = fake_iter
    mocker.patch("grizzly.reduce.core.STRATEGIES", new={"fake": fake_strat})

    target = mocker.Mock(spec_set=Target)
    target.filtered_environ.return_value = {}
    target.asset_mgr = mocker.Mock(spec_set=AssetManager)

    with ReduceManager(
        set(),
        mocker.Mock(spec_set=Sapphire, timeout=30),
        target,
        [testcase],
        ["fake"],
        log_path,
        use_analysis=False,
    ) as mgr:
        with raises(KeyboardInterrupt):
            mgr.run()

    n_reports = 1
    reports = {"20"}
    assert replayer.run.call_count == 30
    expected_dirs = {log_path / "reports"}
    assert set(log_path.iterdir()) == expected_dirs
    tests = {test.read_text() for test in log_path.glob("reports/*-0/test.html")}
    assert tests == reports
    assert sum(1 for _ in (log_path / "reports").iterdir()) == n_reports * 2, list(
        (log_path / "reports").iterdir()
    )


def test_quality_update(mocker, tmp_path):
    """test that the final result gets changed to REDUCED with --fuzzmanager"""
    mocker.patch("grizzly.reduce.strategies.lithium._contains_dd", return_value=True)
    replayer = mocker.patch("grizzly.reduce.core.ReplayManager", autospec=True)
    replayer = replayer.return_value

    def replay_run(testcases, _time_limit, **kw):
        kw["on_iteration_cb"]()
        for test in testcases:
            contents = test["test.html"].read_text()
            if not contents.strip():
                continue
            log_path = tmp_path / f"crash{replayer.run.call_count}_logs"
            log_path.mkdir()
            _fake_save_logs_foo(log_path)
            report = Report(log_path, Path("bin"))
            return [ReplayResult(report, [], True)]
        return []

    replayer.run.side_effect = replay_run

    (tmp_path / "test.html").write_text("123\n")
    testcase = TestCase.load(tmp_path / "test.html")
    assert testcase
    log_path = tmp_path / "logs"

    mocker.patch("grizzly.common.reporter.Collector", autospec=True)
    reporter = mocker.patch("grizzly.reduce.core.FuzzManagerReporter", autospec=True)
    reporter.return_value.submit.return_value = 1234
    update_coll = mocker.patch("grizzly.common.fuzzmanager.Collector")
    target = mocker.Mock(spec_set=Target)
    target.filtered_environ.return_value = {}
    target.asset_mgr = mocker.Mock(spec_set=AssetManager)
    with ReduceManager(
        set(),
        mocker.Mock(spec_set=Sapphire, timeout=30),
        target,
        [testcase],
        ["check", "lines"],
        log_path,
        use_analysis=False,
        report_to_fuzzmanager=True,
    ) as mgr:
        assert mgr.run() == 0

    assert reporter.return_value.submit.call_count == 1
    report_args, _ = reporter.return_value.submit.call_args
    assert len(report_args) == 2
    assert isinstance(report_args[0], list)
    assert len(report_args[0]) == 1
    assert isinstance(report_args[0][0], TestCase)
    assert isinstance(report_args[1], Report)
    assert update_coll.call_count == 1
    assert update_coll.return_value.patch.call_count == 1
    assert update_coll.return_value.patch.call_args[1] == {
        "data": {"testcase_quality": Quality.REDUCED},
    }


def test_include_assets_and_environ(mocker, tmp_path):
    """test report with assets and environment variables"""
    mocker.patch("grizzly.reduce.strategies.lithium._contains_dd", return_value=True)
    replayer = mocker.patch("grizzly.reduce.core.ReplayManager", autospec=True)
    replayer = replayer.return_value

    def replay_run(testcases, _time_limit, **kw):
        kw["on_iteration_cb"]()
        for test in testcases:
            contents = test["test.html"].read_text()
            if not contents.strip():
                continue
            log_path = tmp_path / f"crash{replayer.run.call_count}_logs"
            log_path.mkdir()
            _fake_save_logs_foo(log_path)
            report = Report(log_path, Path("bin"))
            return [ReplayResult(report, [], True)]
        return []

    replayer.run.side_effect = replay_run

    (tmp_path / "test.html").write_text("123\n")
    testcase = TestCase.load(tmp_path / "test.html")
    assert testcase
    log_path = tmp_path / "logs"

    reporter = mocker.patch("grizzly.reduce.core.FilesystemReporter", autospec=True)

    # pylint: disable=unused-argument
    def submit(test_cases, report, force=False):
        assert test_cases
        assert isinstance(report, Report)
        for test in test_cases:
            assert test.assets.get("example")
            assert test.env_vars == {"test": "abc"}

    reporter.return_value.submit.side_effect = submit

    target = mocker.Mock(spec_set=Target)
    target.filtered_environ.return_value = {"test": "abc"}
    with AssetManager(base_path=tmp_path) as asset_mgr:
        (tmp_path / "example_asset").touch()
        asset_mgr.add("example", tmp_path / "example_asset", copy=False)
        target.asset_mgr = asset_mgr
        with ReduceManager(
            set(),
            mocker.Mock(spec_set=Sapphire, timeout=30),
            target,
            [testcase],
            ["check", "lines"],
            log_path,
            use_analysis=False,
        ) as mgr:
            assert mgr.run() == 0

    assert reporter.return_value.submit.call_count == 1


TimeoutTestParams = namedtuple(
    "TimeoutTestParams",
    "durations, interesting, static_timeout, idle_input, idle_output, iter_input,"
    "iter_output, result",
)


@mark.parametrize(
    TimeoutTestParams._fields,
    [
        # 0 duration test sets both timeouts to minimum
        TimeoutTestParams(
            durations=[0],
            interesting=True,
            static_timeout=False,
            idle_input=30,
            idle_output=10,
            iter_input=60,
            iter_output=10,
            result=0,
        ),
        # max duration is used
        TimeoutTestParams(
            durations=[0, 30],
            interesting=True,
            static_timeout=False,
            idle_input=30,
            idle_output=30,
            iter_input=60,
            iter_output=60,
            result=0,
        ),
        # static timeout doesn't affect timeouts
        TimeoutTestParams(
            durations=[0],
            interesting=True,
            static_timeout=True,
            idle_input=30,
            idle_output=30,
            iter_input=60,
            iter_output=60,
            result=0,
        ),
        # uninteresting result doesn't affect timeouts
        TimeoutTestParams(
            durations=[0],
            interesting=False,
            static_timeout=True,
            idle_input=30,
            idle_output=30,
            iter_input=60,
            iter_output=60,
            result=NotReproducible,
        ),
        # test duration affects timeouts
        TimeoutTestParams(
            durations=[10],
            interesting=True,
            static_timeout=False,
            idle_input=30,
            idle_output=15,
            iter_input=60,
            iter_output=20,
            result=0,
        ),
    ],
)
def test_timeout_update(
    mocker,
    tmp_path,
    durations,
    interesting,
    static_timeout,
    idle_input,
    idle_output,
    iter_input,
    iter_output,
    result,
):
    "timeout will be updated based on time to crash"
    mocker.patch("grizzly.reduce.strategies.lithium._contains_dd", return_value=True)
    replayer = mocker.patch("grizzly.reduce.core.ReplayManager", autospec=True)
    replayer = replayer.return_value

    def replay_run(_testcases, _time_limit, **kw):
        kw["on_iteration_cb"]()
        LOG.debug("interesting true")
        log_path = tmp_path / f"crash{replayer.run.call_count}_logs"
        log_path.mkdir()
        _fake_save_logs_foo(log_path)
        report = Report(log_path, Path("bin"))
        return [ReplayResult(report, durations, interesting)]

    replayer.run.side_effect = replay_run

    (tmp_path / "test.html").touch()
    test = TestCase.load(tmp_path / "test.html")
    log_path = tmp_path / "logs"

    target = mocker.Mock(spec_set=Target)
    target.filtered_environ.return_value = {}
    target.asset_mgr = mocker.Mock(spec_set=AssetManager)
    server = mocker.Mock(spec_set=Sapphire, timeout=iter_input)
    with ReduceManager(
        set(),
        server,
        target,
        [test],
        ["check"],
        log_path,
        use_analysis=False,
        idle_delay=idle_input,
        static_timeout=static_timeout,
    ) as mgr:
        if isinstance(result, type) and issubclass(result, BaseException):
            with raises(result):
                mgr.run()
        else:
            assert mgr.run() == result

    assert server.timeout == iter_output
    assert mgr._idle_delay == idle_output



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from __future__ import annotations

from sapphire.server_map import ServerMap

from .storage import TestCase

__all__ = ("IOManager",)
__author__ = "Tyson Smith"
__credits__ = ["Tyson Smith"]


class IOManager:
    __slots__ = (
        "_generated",
        "_report_size",
        "_test",
        "server_map",
        "tests",
    )

    def __init__(self, report_size: int = 1) -> None:
        assert report_size > 0
        self.server_map = ServerMap()
        # tests will be ordered oldest to newest
        self.tests: list[TestCase] = []
        # total number of test cases generated
        self._generated = 0
        self._report_size = report_size
        self._test: TestCase | None = None

    def __enter__(self) -> IOManager:
        return self

    def __exit__(self, *exc: object) -> None:
        self.cleanup()

    def cleanup(self) -> None:
        self.purge()

    def commit(self) -> None:
        assert self._test is not None
        self.tests.append(self._test)
        self._test = None
        # manage testcase cache size
        if len(self.tests) > self._report_size:
            self.tests.pop(0).cleanup()

    def create_testcase(self, adapter_name: str) -> TestCase:
        assert self._test is None
        self._test = TestCase(self.page_name(), adapter_name)
        # reset redirect map
        self.server_map.set_redirect(
            "grz_current_test", self.page_name(), required=False
        )
        self.server_map.set_redirect("grz_next_test", self.page_name(offset=1))
        self._generated += 1
        return self._test

    def page_name(self, offset: int = 0) -> str:
        return f"test_{self._generated + offset:04d}.html"

    def purge(self) -> None:
        if self._test is not None:
            self._test.cleanup()
            self._test = None
        for testcase in self.tests:
            testcase.cleanup()
        self.tests.clear()
